import os
from PySide import QtGui, QtCore
from xicam.plugins import base
from xicam import config
from pyqtgraph import parametertree as pt
from fabio import tifimage
from pipeline import loader, hig, msg
import pyqtgraph as pg
import numpy as np
import subprocess
import xicam.RmcView as rmc
import time
from xicam import threads
from daemon.daemon import daemon
import multiprocessing
import Queue
import glob

# TODO: Remove LUT bar from RMC Timeline (its binary anyways)
# TODO: Check mask behavior


# drag/drop taken from tomography plugin
import platform
op_sys = platform.system()

class plugin(base.plugin):

    """
    HipRMC plugin class - centers images, calls HipRMC as subprocess, then displays output

    NOTE: Running HipRMC assumes that there is a HipRMC folder containing the executable in the working
          directory. If this is not the case, you will need to change the path in inOutViewer.run_RMC where
          HipRMC is called as an executable.


    Attributes
    ----------
    centerwidget : QtGui.QTabWidget
        Standard centerwidget overriding base.plugin. QTabWidget that holds instances of inOutviewer for images and
        HipRMC output
    threadWorker : threads.Worker
        Contains qtCore.QThreadpool on which to run HipRMC jobs, and a queue to hold the jobs


    Parameters
    ----------
    args
        Additional arguments. Not really used
    kwargs
        Additional keyword arguments. Not really used
    """

    name = "HipRMC"

    #center widget is something, rightwidget is none, and leftwidget inherits the default from parent base.plugin
    def __init__(self, *args, **kwargs):


        self.centerwidget = QtGui.QTabWidget()
        self.centerwidget.setDocumentMode(True)
        self.centerwidget.setTabsClosable(True)
        self.centerwidget.tabCloseRequested.connect(self.tabClose)
        self.rightwidget = None

        self.threadWorker = threads.Worker(Queue.Queue())
        self.threadWorker.pool.setExpiryTimeout(1)
        # DRAG-DROP
        self.centerwidget.setAcceptDrops(True)
        self.centerwidget.dragEnterEvent = self.dragEnterEvent
        self.centerwidget.dropEvent = self.dropEvent

        super(plugin, self).__init__(*args, **kwargs)




    def dropEvent(self, e):
        for url in e.mimeData().urls():
            if op_sys == 'Darwin':
                fname = str(NSURL.URLWithString_(str(url.toString())).filePathURL().path())
            else:
                fname = str(url.toLocalFile())
            if os.path.isfile(fname):
                self.openfiles([fname])
            e.accept()

    def dragEnterEvent(self, e):
        e.accept()


    def openfiles(self, paths):

        """
        Overrides inherited 'openfiles' method. Used for opening single image
        """
        self.activate()
        view_widget = inOutViewer(paths = paths, worker = self.threadWorker)
        self.centerwidget.addTab(view_widget, os.path.basename(paths[0]))
        self.centerwidget.setCurrentWidget(view_widget)
        view_widget.drawCameraLocation(view_widget.orig_view, view_widget.cameraLocation)

    def opendirectory(self, folder, operation=None):
        """
        Overrides inherited 'opendirectory' method. Used for opening hiprmc output folders.
        """
        self.activate()
        if type(folder) is list:
            folder = folder[0]
        view_widget = inOutViewer(None, self.threadWorker,)
        self.centerwidget.addTab(view_widget, os.path.basename(folder))
        self.centerwidget.setCurrentWidget(view_widget)

        # check for input image and load into plugin if it exists
        input = glob.glob(os.path.join(folder, 'input_image.tif'))
        if input:
            view_widget.orig_image = np.transpose(loader.loadimage(input[0]))
            if len(view_widget.orig_image.shape) > 2: # gets ride of extra dimensions if there are any
                view_widget.orig_image = np.transpose(view_widget.orig_image).swapaxes(0,1)
                while len(view_widget.orig_image.shape) > 2:
                    view_widget.orig_image = view_widget.orig_image[:,:,0]
            view_widget.orig_view.setImage(view_widget.orig_image)
            view_widget.orig_view.autoRange()

            view_widget.drawROI(0, 0, view_widget.orig_image.shape[0], view_widget.orig_image.shape[1], 'r',
                         view_widget.orig_view.getImageItem().getViewBox())

        view_widget.rmc_view = rmc.rmcView(folder)
        view_widget.rmc_view.findChild(QtGui.QTabBar).hide()
        view_widget.rmc_view.setContentsMargins(0, 0, 0, 0)
        view_widget.image_holder.addWidget(view_widget.rmc_view)

        view_widget.fft_view = rmc.fftView()
        view_widget.fft_view.open_from_rmcView(view_widget.rmc_view.image_list)
        view_widget.fft_view.setContentsMargins(0, 0, 0, 0)
        view_widget.image_holder.addWidget(view_widget.fft_view)

        view_widget.image_holder.setCurrentIndex(2)


    def tabClose(self,index):
        self.centerwidget.widget(index).deleteLater()

class LogViewer(pg.ImageView):
    """
    Class to view images with log intensity
    """

    def setImage(self,*args,**kwargs):
        super(LogViewer, self).setImage(*args,**kwargs)

        # levelmin = np.log(self.levelMin)/np.log(1.5)
        # levelmax = np.log(self.levelMax)/np.log(1.5)
        levelmin = np.log(self.levelMin)
        levelmax = np.log(self.levelMax)
        if np.isnan(levelmin): levelmin = 0
        if np.isnan(levelmax): levelmax = 1
        if np.isinf(levelmin): levelmin = 0

        self.ui.histogram.setLevels(levelmin, levelmax)



class inOutViewer(QtGui.QWidget, ):
    def __init__(self, paths, worker, parent=None):
        """
        Class that holds image to be processed by HipRMC, image after it has been centered, and HipRMC output

        Attributes
        ----------
        emitter : threads.Emitter
            Holds and emits a signal when fHipRMc done processing
        interrupt : bool
            flag - set true if rmc processing was interrupted; affects post-rmc processes
        cameraLocation : tuple
            2-tuple (x,y) of camera location on input image
        rmcView : ximcam.RmcView.rncView
            Timeline viewer which holds and displays HipRMC output
        orig_image: np.array
            Original input image
        edited_image : np.array
            Image with camera location adjusted to its center
        orig_view : pyqtgraph.ImageView
            Holds the original image
        edited_view : pyqtgraph.ImageView
            Holds the image after camera location adjustment
        image_holder : QtGui.StackedWidget
            Main widget of plugin. Holds original and edited images, as well as HipRMC output, in tabs
        scatteringParams : pyqtgraph.parametertree
            Occupies right side of main widget. Holds configparams
        configparams : pyqtgraph.Parameter
            Class held by scatteringParams which holds parameter values for HipRMC
        output, err : str
            Output and error from HipRMC subprocess call
        headings : QtGui.QTabBar
            Displays name of corresponding tab of image_holder

        Parameters
        ----------
        paths : str/list of str
            Path to input dataset
        worker: threads.Worker
            Worker which queues up jobs and runs them on a QtCore.QThreadpool
        parent : QtGui.QWidget
            parent widget
        args
            Additional arguments
        kwargs
            Additional keyword arguments
        """


        super(inOutViewer, self).__init__(parent=parent)

        self.emitter = threads.Emitter()
        self.interrupt = False
        self.cameraLocation = config.activeExperiment.center
        self.rmc_view= None
        self.edited_image = None
        self.worker = worker


        # holders for original and edited images
        self.orig_view = LogViewer()
        self.orig_view.setContentsMargins(0,0,0,0)
        self.edited_view = LogViewer()
        self.edited_view.setContentsMargins(0,0,0,0)

        if type(paths) == list:
            self.path = paths[0]
        else:
            self.path = paths

        self.image_holder = QtGui.QStackedWidget()
        self.image_holder.setContentsMargins(0,0,0,0)

        # configuring right widget
        sideWidget = QtGui.QWidget()
        sideWidgetFormat = QtGui.QVBoxLayout()
        sideWidgetFormat.setContentsMargins(0, 0, 0, 0)

        # if paths is None, inOutViewer will only hold HipRMC output and the images/parameter table are not necessary
        if paths is not None:

            self.orig_image = np.transpose(loader.loadimage(self.path))
            if len(self.orig_image.shape) > 2:
                self.orig_image = np.transpose(self.orig_image).swapaxes(0,1)
                while len(self.orig_image.shape) > 2:
                    self.orig_image = self.orig_image[:,:,0]
            self.orig_view.setImage(self.orig_image)
            self.orig_view.autoRange()
            try:
                start_size = max(self.orig_image.shape)/10
            except ValueError:
                msg.showMessage("Image must be 2-D")

            scatteringHolder = QtGui.QStackedWidget()

            image_name = self.path.split('/')[-1].split('.')[0]
            self.scatteringParams = pt.ParameterTree()
            params = [{'name': 'Num tiles', 'type': 'int', 'value': 1, 'default': 1},
                      {'name': 'Loading factor', 'type': 'float', 'value': 0.5, 'default': 0.5},
                      {'name': 'Scale factor', 'type': 'int', 'value': 32, 'default': 32},
                      {'name': 'Numsteps factor', 'type': 'int', 'value': 100, 'default': 100},
                      {'name': 'Model start size', 'type': 'int', 'value': start_size},
                      {'name': 'Save name', 'type': 'str', 'value': 'hiprmc_' + image_name},
                      {'name': 'Mask image', 'type': 'str'}]
            self.configparams = pt.Parameter.create(name='Configuration', type='group', children=params)
            self.scatteringParams.setParameters(self.configparams, showTop=False)
            scatteringHolder.addWidget(self.scatteringParams)

            # # is there a better way to check for correct dimensions?
            # if len(self.orig_image.shape) > 2:
            #     shape = (self.orig_image.shape[1], self.orig_image.shape[2])
            # else:
            #     shape = self.orig_image.shape

            self.drawROI(0, 0, self.orig_image.shape[0], self.orig_image.shape[1], 'r',
                         self.orig_view.getImageItem().getViewBox())

            scatteringHolder.setFixedHeight(300)
            sideWidgetFormat.addWidget(scatteringHolder)

        centerButton = QtGui.QPushButton("Center camera location")
        runButton = QtGui.QPushButton("Run RMC processing")
        stopButton = QtGui.QPushButton("Stop RMC")
        sideWidgetFormat.addSpacing(5)
        sideWidgetFormat.addWidget(centerButton)
        sideWidgetFormat.addSpacing(5)
        sideWidgetFormat.addWidget(runButton)
        sideWidgetFormat.addSpacing(5)
        sideWidgetFormat.addWidget(stopButton)
        sideWidgetFormat.addStretch(10)
        sideWidget.setLayout(sideWidgetFormat)

        # connect buttons to processing
        centerButton.clicked.connect(self.center)
        runButton.clicked.connect(self.runRMC)
        stopButton.clicked.connect(self.stop_threads)

        # tab headings for main widget
        self.headings = QtGui.QTabBar(self)
        self.headings.addTab('Original Image')
        self.headings.addTab('Recentered Image')
        self.headings.addTab('RMC Timeline')
        self.headings.addTab('FFT RMC Timeline')
        self.headings.setShape(QtGui.QTabBar.TriangularSouth)


        leftWidget = QtGui.QWidget()
        sidelayout = QtGui.QVBoxLayout()
        sidelayout.addWidget(self.image_holder)
        sidelayout.addWidget(self.headings)
        leftWidget.setLayout(sidelayout)

        fullPlugin = QtGui.QSplitter()
        fullPlugin.addWidget(leftWidget)
        fullPlugin.addWidget(sideWidget)

        h = QtGui.QHBoxLayout()
        h.setContentsMargins(0, 0, 0, 0)
        h.addWidget(fullPlugin)
        self.setLayout(h)

        self.image_holder.addWidget(self.orig_view)
        self.image_holder.addWidget(self.edited_view)

        self.headings.currentChanged.connect(self.currentChanged)
        self.image_holder.currentChanged.connect(self.headings.setCurrentIndex)


    def currentChanged(self, index):
        """
        Slot to receive centerwidget's currentchanged signal when a new tab is selected
        """
        if self.image_holder.widget(index):
            self.image_holder.setCurrentIndex(index)
        else:
            self.headings.setCurrentIndex(self.image_holder.currentIndex())


    def center(self, sample=True):
        """
        Slot to receive signal when user requests a centered image. Performs centering and writes new image into
        current working directory

        Parameters
        ----------
        sample: boolean, optional
            flag whether image to be centered is the original sample or the mask image
        """

        if sample:
            if self.edited_image is not None:
                msg.showMessage('Image already centered.')
                return

            image = self.orig_image
        else:
            image = self.mask

        #resize image so that it's in center and displays output if a sample image
        xdim = int(image.shape[0])
        ydim = int(image.shape[1])

        newx = xdim + 2*abs(self.cameraLocation[0]-xdim/2)
        newy = ydim + 2*abs(self.cameraLocation[1]-ydim/2)
        self.new_dim = max(newx,newy)

        self.edited_image = np.ones((self.new_dim,self.new_dim),dtype = np.int)
        new_center = (self.new_dim/2,self.new_dim/2)

        lowleft_corner_x = int(new_center[0]-self.cameraLocation[0])
        lowleft_corner_y = int(new_center[1]-self.cameraLocation[1])

        self.edited_image[lowleft_corner_x:lowleft_corner_x+xdim,lowleft_corner_y: lowleft_corner_y+ydim] = image

        # save image
        if sample:
            self.write_path = self.path

            if self.write_path.endswith('.tif'):
                self.write_path = self.write_path[:-4] + '_centered.tif'
            else:
                self.write_path += '_centered.tif'
            self.write_path_sample = self.write_path

            img = tifimage.tifimage(np.rot90((self.edited_image.astype(float) /
                                              self.edited_image.max() * 2 ** 16).astype(np.int16)))
        else:
            self.write_path = self.mask_path

            if self.write_path.endswith('.tif'):
                self.write_path = self.write_path[:-4] + '_centered.tif'
            else:
                self.write_path += '_centered.tif'
            self.write_path_mask = self.write_path

            img = tifimage.tifimage(np.rot90(self.edited_image.astype(float)))




        img.write(self.write_path)

        if sample:
            self.edited_view.setImage(self.edited_image)

            box = self.drawCameraLocation(self.edited_view,new_center)
            self.drawROI(lowleft_corner_x,lowleft_corner_y,xdim,ydim,'r', box)
            self.drawROI(0,0,self.new_dim,self.new_dim, 'b', box)

            # this is a temporary fix for a problem: pushing a button changes tab back to first
            self.image_holder.setCurrentIndex(1)

    def drawCameraLocation(self,box,location):
        """
        Draws camera location as dot on image

        Parameters
        ----------
        box : pyqtgraph.ImageView
            Class which holds the dot drawn by the function
        location: tuple
            2-tuple of camera location
        """


        cameraBox = box.getImageItem().getViewBox()
        cameraPlot = pg.ScatterPlotItem()
        cameraBox.addItem(cameraPlot)
        cameraPlot.setData([location[0]], [location[1]], pen=None,
                                symbol='o' , brush=pg.mkBrush('#FFA500'))

        return cameraBox


    def drawROI(self, xpos, ypos, xdim,ydim, color, view_box):
        """
        Draws camera location as dot on image

        Parameters
        ----------
        xpos, ypos: int
            Location of lower left corner of ROI
        xdim, ydim : int
            Size of ROI in x and y dimensions
        color : str
            Color of ROI
        view_box : pyqtgraph.ImageView
            Class which holds the ROI drawn by the function
        """

        roi = pg.RectROI((xpos,ypos),(xdim,ydim),movable = False,removable=True, snapSize = 100000000, scaleSnap=True)
        roi.setPen(color = color)

        view_box.addItem(roi)


    def runRMC(self):
        """
        Slot to receive signal when user requests HipRMC calculation. Writes hig file of parameter values and
        calls HipRMC as subprocess
        """


        msg.showMessage('Running RMC for centered version of {}'.format(self.path), timeout=0)

        if self.rmc_view is not None:
            self.image_holder.removeWidget(self.rmc_view)

        if self.edited_image is None:
            msg.showMessage('Error: must center image before running HipRMC',timeout = 0)
            msg.clearMessage()
            return



        params = self.configparams

        hig_info = {'hipRMCInput': {'instrumentation': {'inputimage': "{}".format(self.write_path_sample),
                                             'imagesize': [self.new_dim, self.new_dim ],
                                             'numtiles': params.child('Num tiles').value(),
                                             'loadingfactors': [params.child('Loading factor').value()]},
                         'computation': {'runname': "{}".format(params.child('Save name').value()),
                                         'modelstartsize': [params.child('Model start size').value(),
                                                            params.child('Model start size').value()],
                                         'numstepsfactor': params.child('Numsteps factor').value(),
                                         'scalefactor': params.child('Scale factor').value()}}}

        self.mask_path = params.child('Mask image').value()
        if self.mask_path and self.mask_path != "None":
            self.mask = np.transpose(loader.loadimage(self.mask_path))
            self.center(False)
            hig_info['hipRMCInput']['instrumentation']['maskimage'] = "{}".format(self.write_path_mask)

        h = hig.hig(**hig_info)
        self.hig_name = os.path.join(os.path.abspath('.'), params.child('Save name').value())

        if not self.hig_name.endswith('.hig'):
            self.hig_name += '.hig'

        # write hig file to disk
        h.write(self.hig_name)
        self.save_name = params.child('Save name').value()
        self.start_time = time.time()

        # starts filewatcher to watch for new hiprmc folder, and the HipRMC job
        # also starts worker if it is not already running
        process = threads.RunnableMethod(method = self.run_RMCthread, finished_slot = self.RMC_done,
                                         except_slot=self.hiprmc_not_found)
        self.file_watcher = NewFolderWatcher(path=os.path.abspath("."), experiment=None)

        # when filewatcher gets rmc folder, it passes it to self.start_watcher to start another watcher
        self.file_watcher.sigFinished.connect(self.start_watcher)
        watcher = threads.RunnableMethod(method=self.file_watcher.run,)
        self.worker.queue.put(watcher)
        self.worker.queue.put(process)

        if not self.worker.isRunning():
            self.worker.start()


    def run_RMCthread(self):
        """
        Slot to receive signal to run HipRMC as subprocess on background thread
        """
        if os.path.isfile('./hiprmc/bin/hiprmc'):
            self.proc = subprocess.Popen(['./hiprmc/bin/hiprmc', self.hig_name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            self.output, self.err = self.proc.communicate()
        else:
            raise Exception

    def hiprmc_not_found(self):
        """
        Slot to receive exception signal in case hiprmc is not found.
        """
        msg.showMessage('Cannot find HipRMC executable. Cannot run HipRMC.')
        QtGui.QMessageBox.critical(self, 'Error', 'Cannot find HipRMC executable. Cannot run HipRMC.')


    def start_watcher(self, folder):
        """
        Slot to receive signal to start looking for hiprmc output in 'folder'

        Parameters
        ----------
        folder : str, unicode
            folder to watch for hiprmc output
        """

        self.rmc_folder = folder

        # create and add rmcView to main plugin widget
        self.rmc_view = rmc.rmcView(self.rmc_folder)
        self.rmc_view.findChild(QtGui.QTabBar).hide()
        self.rmc_view.setContentsMargins(0, 0, 0, 0)
        self.image_holder.addWidget(self.rmc_view)

        # fftview to view fft transforms of output images
        self.fft_view = rmc.fftView()
        self.fft_view.open_from_rmcView(self.rmc_view.image_list)
        self.fft_view.setContentsMargins(0, 0, 0, 0)
        self.image_holder.addWidget(self.fft_view)

        # starts another filewatcher to watch rmc folder and add images to rmc_view as they are output
        self.rmc_watcher = HipRMCWatcher(path=self.rmc_folder,experiment=self.rmc_view)
        self.emitter.sigFinished.connect(self.rmc_watcher.stop)
        watchRMC = threads.RunnableMethod(method=self.rmc_watcher.run,)
        self.worker.queue.put(watchRMC)
        self.rmc_watcher.sigCallback.connect(self.add_images)
        if self.rmc_view.image_list:
           self.image_holder.setCurrentIndex(2)

        if not self.worker.isRunning():
            self.worker.start()

    def add_images(self, root):
        """
        Looks in file given by 'root' and adds images to rmcView and fftView

        Parameters
        ----------
        root : str, unicode
            path to check for images to add to rmcView and fftView
        """
        if not self.rmc_view.image_list:
            self.image_holder.setCurrentIndex(2)
        self.rmc_view.addNewImages(root=root)
        self.fft_view.open_from_rmcView(self.rmc_view.image_list)


    def stop_threads(self):
        """
        Stops all background threads (worker, any filewatcher, hiprmc
        """
        try:
            self.worker.stop()
            self.file_watcher.stop()
            self.rmc_watcher.stop()
            self.interrupt = True
            self.proc.terminate()
        except IOError:
            pass

    @QtCore.Slot()
    def RMC_done(self):
        """
        Slot to receive signal when HipRMC calculation is done. Emits signal to main thread to load output
        """
        run_time = time.time() - self.start_time

        if not self.interrupt:
            os.rename(self.hig_name, '{}/{}.hig'.format(self.rmc_folder, self.save_name))

            # write output of RMC to file in hiprmc output folder
            output_path = self.rmc_folder + "/{}_rmc_output.txt".format(self.save_name)
            with open(output_path, 'w') as txt:
                txt.write(self.output)

            msg.showMessage('HipRMC complete. Run time: {:.2f} s'.format(run_time))
            self.emitter.sigFinished.emit()
        else:
            try:
                os.rename(self.hig_name, '{}/{}.hig'.format(self.rmc_folder, self.save_name))
            except OSError:
                pass

            msg.showMessage('HipRMC interrupted by user. Run time: {:.2f} s'.format(run_time))

        self.interrupt = False


class NewFolderWatcher(daemon):

    """
    Daemon subclass to watch for hiprmc folder when it is created
    """

    sigFinished = QtCore.Signal(str)

    def process(self, path, files):

        if files:
            folder = path + '/' + files[0]
            self.sigFinished.emit(folder)
            self.stop()

    def run(self):

        if self.procold:
            self.process(self.path, self.childfiles)

        try:
            while not self.exiting:
                time.sleep(.1)
                self.checkdirectory()  # Force update; should not have to do this -.-
        except KeyboardInterrupt:
            pass

    def stop(self):
        self.exiting = True
        print ("thread stop - %s" % self.exiting)

    def __del__(self):
        self.exiting = True
        self.wait()


    def checkdirectory(self):
        """
        Checks a directory for new files, comparing what files are there now vs. before
        """
        updatedchildren = set(os.listdir(self.path))
        newchildren = updatedchildren - self.childfiles
        self.childfiles = updatedchildren
        self.process(self.path, list(newchildren))


class HipRMCWatcher(daemon):

    """
    Daemon subclass to watch hiprmc output folder. Emits sigCallback whenever folder contents change
    """

    num_cores = multiprocessing.cpu_count()
    sigCallback = QtCore.Signal(str)

    def run(self):

        if self.procold:
            self.process(self.path, self.childfiles)

        try:
            while not self.exiting:
                time.sleep(.1)
                self.checkdirectory()  # Force update; should not have to do this -.-
        except KeyboardInterrupt:
            pass

    def checkdirectory(self):
        """
        Checks a directory for new files, comparing what files are there now vs. before
        """
        updatedchildren = set(os.listdir(self.path))
        newchildren = updatedchildren - self.childfiles
        self.childfiles = updatedchildren
        self.process(self.path, list(newchildren))


    def process(self, path, files):
        if files:
            self.sigCallback.emit(path)






import paramiko


class SSHClient(paramiko.SSHClient):
    """Save a Paramiko SSH Connection"""

    def __init__(self, username,host,password):
        port = 22
        self.host = host

        super(SSHClient, self).__init__()
        self.load_system_host_keys()
        self.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.connect(host, port, username, password)




import platform
from pipeline import msg

# Use NSURL as a workaround to pyside/Qt4 behaviour for dragging and dropping on OSx
op_sys = platform.system()
if op_sys == 'Darwin':
    try:
        from Foundation import NSURL
    except ImportError:
        msg.logMessage('NSURL not found. Drag and drop may not work correctly',msg.WARNING)


import base, viewer
from PySide import QtGui,QtCore
import os
# from moviepy.editor import VideoClip
import numpy as np

import widgets
from pipeline import calibration
from xicam.widgets.NDTimelinePlotWidget import TimelinePlot
from collections import OrderedDict


class TimelinePlugin(base.plugin):  ##### Inherit viewer instead!!!
    name = 'Timeline'
    sigUpdateExperiment = viewer.ViewerPlugin.sigUpdateExperiment

    def __init__(self, *args, **kwargs):
        self.centerwidget = QtGui.QTabWidget()
        self.centerwidget.currentChanged.connect(self.currentChanged)
        self.centerwidget.setDocumentMode(True)
        self.centerwidget.setTabsClosable(True)
        self.centerwidget.tabCloseRequested.connect(self.tabCloseRequested)

        self.bottomwidget = TimelinePlot()


        # Share right modes with viewer
        self.rightmodes = viewer.rightmodes

        self.toolbar = widgets.toolbar.difftoolbar()
        self.toolbar.connecttriggers(self.calibrate, self.centerfind, self.refinecenter, self.redrawcurrent,
                                     self.redrawcurrent, self.remeshmode, self.linecut, self.vertcut,
                                     self.horzcut, self.redrawcurrent, self.redrawcurrent, self.redrawcurrent,
                                     self.roi, self.arccut, self.polymask, process=self.process)
        super(TimelinePlugin, self).__init__(*args, **kwargs)

        # self.booltoolbar.actionTimeline.triggered.connect(self.openSelected)

        # DRAG-DROP
        self.centerwidget.setAcceptDrops(True)
        self.centerwidget.dragEnterEvent = self.dragEnterEvent
        self.centerwidget.dropEvent = self.dropEvent

    def dragEnterEvent(self, e):
        e.accept()
        # TODO: We should do something a bit less aggressive here!

    def dropEvent(self, e):
        if op_sys == 'Darwin':
            fnames = [str(NSURL.URLWithString_(str(url.toString())).filePathURL().path()) for url in
                      e.mimeData().urls()]
        else:
            fnames = e.mimeData().urls()
        e.accept()
        self.openfiles(fnames)



    def tabCloseRequested(self, index):
        self.centerwidget.widget(index).deleteLater()

    def getCurrentTab(self):
        return self.centerwidget.currentWidget().widget

    def currentImage(self):
        return self.getCurrentTab()

    def calibrate(self, algorithm=calibration.fourierAutocorrelation, calibrant='AgBh'):
        self.getCurrentTab().calibrate(algorithm, calibrant)

    def centerfind(self):
        self.getCurrentTab().centerfind()

    def refinecenter(self):
        self.getCurrentTab().refinecenter()

    def redrawcurrent(self):
        self.getCurrentTab().redrawimage()

    def remeshmode(self):
        self.getCurrentTab().redrawimage()
        self.getCurrentTab().replot()

    def linecut(self):
        self.getCurrentTab().linecut()

    def vertcut(self):
        self.getCurrentTab().verticalcut()

    def horzcut(self):
        self.getCurrentTab().horizontalcut()

    def roi(self):
        self.getCurrentTab().roi()

    def arccut(self):
        self.getCurrentTab().arccut()

    def polymask(self):
        self.getCurrentTab().polymask()

    def process(self):
        self.getCurrentTab().processtimeline()

    def makeVideo(self):
        pass  # disabled until solution is found for distributable version

        # fps, ok = QtGui.QInputDialog.getDouble(self.centerwidget, u'Enter frames per second:', u'Enter fps', value=24)
        #
        # def make_frame(t):
        # """ returns an image of the frame at time t """
        # # ... create the frame with any library
        #     img = self.getCurrentTab().simg[int(t * fps)].data
        #     img = np.rot90((np.log(img * (img > 0) + (img < 1))), 1)
        #     img = convertto8bit(img)
        #     return np.asarray(np.dstack((img, img, img)), dtype=np.uint8)
        #
        # if fps and ok:
        #     animation = VideoClip(make_frame, duration=len(self.getCurrentTab().simg) / fps)
        #
        #     # For the export, many options/formats/optimizations are supported
        #
        #     path, ok = QtGui.QFileDialog.getSaveFileName(self.centerwidget, 'Save Video', os.path.splitext(self.getCurrentTab().simg[0].filepath)[0] + '.mp4','*.mp4')
        #
        #     if path and ok:
        #         if os.path.splitext(path)[-1] == '.mp4':
        #             animation.write_videofile(path, fps=fps)  # export as video
        #         elif os.path.splitext(path)[-1] == '.gif':
        #             animation.write_gif(path, fps=fps)  # export as GIF (slow)
        #         else:
        #             animation.write_videofile(path, fps=fps)  # export as video
        #             print 'Error: Unrecognized extension...'


    def currentChanged(self, index):
        for tab in [self.centerwidget.widget(i) for i in range(self.centerwidget.count())]:
            tab.unload()
        self.centerwidget.currentWidget().load()

    @QtCore.Slot(object,object)
    def unpack(self,*args):
        if isinstance(args[1], OrderedDict):
            self.bottomwidget.addData(args[0], **args[1])
        elif isinstance(args[1], tuple):
            self.bottomwidget.addData(args[0], *args[1])
        else:
            self.bottomwidget.addData(args[0], args[1])

    def openfiles(self, files, operation=None, operationname=None):
        self.activate()
        widget = widgets.OOMTabItem(itemclass=widgets.timelineViewer, files=files, toolbar=self.toolbar)
        self.centerwidget.addTab(widget, 'Timeline: ' + os.path.basename(files[0]) + ', ...')
        self.centerwidget.setCurrentWidget(widget)


        self.getCurrentTab().sigAddTimelineData.connect(self.unpack)
        self.getCurrentTab().sigClearTimeline.connect(self.bottomwidget.clear)


def convertto8bit(image):
    image *= (np.iinfo(np.uint8).max - 1) / float(np.max(image))
    return image.astype(np.uint8).copy()



import importlib
from pipeline import msg
import tomopy

# Packages to import
PACKAGE_LIST = ['astra', 'dxchange', 'tomocam', 'pyF3D']

# Dictionary with package names as keys and package objects as values
packages = {}

for name in PACKAGE_LIST:
    try:
        package = importlib.import_module(name)
        packages[name] = package
        msg.logMessage('{} module loaded'.format(name), level=20)
    except ImportError as ex:
        msg.logMessage('{} module not available'.format(name), level=30)  # 30 -> warning'

# Tomopy is actually necessary, so its important that its a hard import
packages['tomopy'] = tomopy

# Add the extra functions
import pipelinefunctions
packages['pipelinefunctions'] = pipelinefunctions

if 'tomocam' in packages:
    import mbir
    packages['mbir'] = mbir


import pyFAI.detectors
import fabio
import logging
from collections import OrderedDict
import numpy as np

class PrincetonMTE(pyFAI.detectors.Detector):
    """
    Princeton Instrument, PI-MTE CCD
    """
    force_pixel = True
    MAX_SHAPE = (2048, 2048)
    BINNED_PIXEL_SIZE = {1: 13.5e-6,
                         2: 27e-6}


    def __init__(self):
        pixel1 = 27e-6
        pixel2 = 27e-6
        super(PrincetonMTE, self).__init__(pixel1=pixel1, pixel2=pixel2)

    def get_binning(self):
        return self._binning

    def set_binning(self, bin_size=(1, 1)):
        """
        Set the "binning" of the detector,
        @param bin_size: set the binning of the detector
        @type bin_size: int or (int, int)
        """
        if "__len__" in dir(bin_size) and len(bin_size) >= 2:
            bin_size = int(round(float(bin_size[0]))), int(round(float(bin_size[1])))
        else:
            b = int(round(float(bin_size)))
            bin_size = (b, b)
        if bin_size != self._binning:
            if (bin_size[0] in self.BINNED_PIXEL_SIZE) and (bin_size[1] in self.BINNED_PIXEL_SIZE):
                self._pixel1 = self.BINNED_PIXEL_SIZE[bin_size[0]]
                self._pixel2 = self.BINNED_PIXEL_SIZE[bin_size[1]]
            else:
                # logger.warning("Binning factor (%sx%s) is not an official value for Princeton Instruments detectors" % (bin_size[0], bin_size[1]))
                self._pixel1 = self.BINNED_PIXEL_SIZE[1] / float(bin_size[0])
                self._pixel2 = self.BINNED_PIXEL_SIZE[1] / float(bin_size[1])
            self._binning = bin_size
            self.shape = (self.max_shape[0] // bin_size[0],
                          self.max_shape[1] // bin_size[1])

    binning = property(get_binning, set_binning)

    def __repr__(self):
        return "Detector %s\t PixelSize= %.3e, %.3e m" % \
               (self.name, self._pixel1, self._pixel2)

    def guess_binning(self, data):
        """
        Guess the binning/mode depending on the image shape
        @param data: 2-tuple with the shape of the image or the image with a .shape attribute.
        """
        if "shape" in dir(data):
            shape = data.shape
        else:
            shape = tuple(data[:2])
        bin1 = self.MAX_SHAPE[0] // shape[0]
        bin2 = self.MAX_SHAPE[1] // shape[1]
        self._binning = (bin1, bin2)
        self.shape = shape
        self.max_shape = shape
        self._pixel1 = self.BINNED_PIXEL_SIZE[bin1]
        self._pixel2 = self.BINNED_PIXEL_SIZE[bin2]
        self._mask = False
        self._mask_crc = None


class PhotonicScience(pyFAI.detectors.Detector):
    """
    Photonic Science CCD Detector

    """
    aliases = ["Photonic Science"]
    force_pixel = True
    MAX_SHAPE = (1042, 1042)
    DEFAULT_PIXEL1 = DEFAULT_PIXEL2 = 200e-6

    def __init__(self, pixel1=200e-6, pixel2=200e-6):
        super(PhotonicScience, self).__init__(pixel1=pixel1, pixel2=pixel2)
        if (pixel1 != self.DEFAULT_PIXEL1) or (pixel2 != self.DEFAULT_PIXEL2):
            self._binning = (int(2 * pixel1 / self.DEFAULT_PIXEL1), int(2 * pixel2 / self.DEFAULT_PIXEL2))
            self.shape = tuple(s // b for s, b in zip(self.max_shape, self._binning))
        else:
            self.shape = (1042/2, 1042/2)
            self._binning = (2, 2)

    def __repr__(self):
        return "Detector %s\t PixelSize= %.3e, %.3e m" % \
            (self.name, self._pixel1, self._pixel2)

class Xpad_BM02(pyFAI.detectors.Xpad_flat):
    MAX_SHAPE = (958, 576)

    aliases = ['Xpad (ESRF BM02)']

    def calc_mask(self):
        """
        Returns a generic mask for Xpad detectors...
        discards the first line and raw form all modules:
        those are 2.5x bigger and often mis - behaving
        """
        if self.max_shape is None:
            raise NotImplementedError("Generic Xpad detector does not"
                                      " know the max size ...")
        mask = np.zeros(self.max_shape, dtype=np.int8)
        # workinng in dim0 = Y
        for i in range(0, self.max_shape[0], self.module_size[0]):
            try:
                mask[i + self.module_size[0]-2, :] = 1
                mask[i + self.module_size[0] - 1, :] = 1
            except IndexError:
                pass
        # workinng in dim1 = X
        for i in range(0, self.max_shape[1], self.module_size[1]):
            try:
                mask[:, i] = 1
                mask[:, i + self.module_size[1] - 1] = 1
            except IndexError:
                pass
        return mask

class Maxipix_BM05(pyFAI.detectors.Maxipix5x1):
    aliases = ['Maxipix (ESRF BM05)']

    def calc_mask(self):
        """
        Returns a generic mask for Mexipix detectors...
        """
        if self.max_shape is None:
            raise NotImplementedError("Generic Pilatus detector does not know "
                                      "its max size ...")
        mask = np.zeros(self.max_shape, dtype=np.int8)
        # workinng in dim0 = Y
        for i in range(self.module_size[0], self.max_shape[0], self.module_size[0] + self.MODULE_GAP[0]):
            mask[i: i + self.MODULE_GAP[0], :] = 1
        # workinng in dim1 = X
        for i in range(self.module_size[1], self.max_shape[1], self.module_size[1] + self.MODULE_GAP[1]):
            mask[:, i - 1: 1 + i + self.MODULE_GAP[1]] = 1
        return mask



ALL_DETECTORS = OrderedDict((cls().name,cls) for cls in sorted(pyFAI.detectors.ALL_DETECTORS.values(),key=lambda cls:cls().name))


# -*- coding: utf-8 -*-
from copy import copy
from pipeline.loader import StackImage

__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque",
               "Holden Parks", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


class ProjectionStack(StackImage):
    """
    Simply subclass of StackImage for Tomography Projection stacks.

    Attributes
    ----------
    flats : ndarray
        Flat field data
    darks : ndarray
        Dark field data
    """

    def     __init__(self, filepath=None, data=None):
        super(ProjectionStack, self).__init__(filepath=filepath, data=data)
        try:
            self.flats = self.fabimage.flats
            self.darks = self.fabimage.darks
        except AttributeError:
            self.flats, self.darks = None, None

    def _getframe(self, frame=None):
        """
        Override base class method to transpose data for viewing
        """

        return super(ProjectionStack, self)._getframe(frame=frame).transpose()




class SinogramStack(StackImage):
    """
    StackImage  subclass for Tomography Sinogram stacks.
    """

    def __init__(self, filepath=None, data=None):
        super(SinogramStack, self).__init__(filepath=filepath, data=data)
        self._cachesize = 10

    def __new__(cls):
        cls.invalidatecache()

    @classmethod
    def cast(cls, obj):
        """
        Use this to cast a ProjectionStack into a SinogramStack

        Parameters
        ----------
        obj : StackImage/ProjectionStack
            Instance object to cast into Sinogram stack
        Returns
        -------
        SinogramStack
            Object cast into SinogramStack
        """

        new_obj = copy(obj)
        new_obj.__class__ = cls
        new_obj.shape = new_obj.shape[2], new_obj.shape[0], new_obj.shape[1]
        return new_obj

    def _getimage(self, frame):
        """
        Override method from base class to read along sinogram dimension
        """
        return self.fabimage[:, frame, :]

    def _getframe(self, frame=None):
        """
        Override base class method to transpose data for viewing
        """

        return super(SinogramStack, self)._getframe(frame=frame).transpose()



import pandas as pd
from modpkgs.memoize import memoized_property
import numpy as np
try:
    from larch import Interpreter
    from larch_plugins.math.mathutils import index_of
    from larch_plugins.xafs import pre_edge
    mylarch = Interpreter()
except ImportError:
    print('Larch is not installed; XAS plugin will be disabled.')

from os import path



def open(f):
    # determine which class to use somehow
    for cls in XASclasses:
        if hasattr(cls,'validate'):
            try:
                cls.validate(f)
            except Exception as ex:
                continue
            else:
                return cls(f)
        try:
            obj=cls(f)
            return obj
        except Exception as ex:
            continue

    raise ValueError('The selected file could not be loaded with any recognized format.')


class XASSpectra(object):
    def __init__(self,f):
        self.scans = []

        self.read(f)


    def read(self, filename, frame=None):
        """
        To be overridden - fill in self.header and self.data
        """
        raise Exception("Class has not implemented read method yet")
        #        return self

    def normalize(self):
        for scan in self.scans:
            scan['I_norm'] = scan.Data.astype(float)/scan.Reference.astype(float)

    def pre_edge(self):
        for scan in self.scans:
            pre_edge(scan.Energy, scan.I_norm, pre1=-10, pre2=-2, group=scan, _larch=mylarch)
            # print 'Automatic edge position found as', scan.e0, 'eV'

            scan.I_norm = scan.I_norm - scan.pre_edge  # Subtract fitted pre-edge curve
            # mdat[str(idx)] = mdat.y  # Save leveled data in matrix for later
            # select data range.

    def lowcut(self, limit=50):
        self.scans = [scan[limit:] for scan in self.scans]


    def treat(self):
        self.normalize()
        self.pre_edge()
        self.lowcut()

XASclasses=[]
def register_xasclass(cls):
    global XASclasses
    XASclasses.append(cls)
    return cls

@register_xasclass
class BL632Spectra(XASSpectra):
    colmap = {'TEY': 'Data', 'I1': 'Reference'}
    def read(self, filename, scan=None):
        self.rawdata = pd.read_table(filename,header=1,names=['Energy','I1','TEY','I0'])
        self.rawdata.rename(columns=self.colmap, inplace=True)
        self.scans=[self.rawdata]
        return self

    is_stacked = False
    num_scans = 1

    @staticmethod
    def validate(filename):
        assert path.splitext(filename)[-1] == '.dat'
        assert not pd.read_table(filename,header=1,names=['Energy','I1','TEY','I0']).empty


    @memoized_property
    def num_bins(self):
        return self.rawdata.shape[0]

    def scan(self, N):
        return self.scans[N]

@register_xasclass
class BL6311Spectra(XASSpectra):
    colmap = {'Counter 3': 'Data', 'Counter 2': 'Reference'}
    skiprows = 10

    def read(self, filename, scan=None):
        """import scan file"""
        scandata_f = pd.read_csv(filename, sep='\t', skiprows=self.skiprows)

        self.rawdata = scandata_f
        self.rawdata.rename(columns = self.colmap, inplace = True)
        self.scans = [self.scan(i) for i in range(self.num_scans)]
        return self

    @staticmethod
    def validate(filename,skiprows=skiprows):
        assert path.splitext(filename)[-1] == '.txt'
        assert not pd.read_csv(filename, sep='\t', skiprows=skiprows).empty

    @memoized_property
    def is_stacked(self):
        """Determine stacked or single scan file(s) and informs user"""
        return len(self.rawdata.Energy) != len(set(self.rawdata.Energy))

    @memoized_property
    def num_scans(self):
        # Number of scans in file using set to find repeating numbers in energy column
        return len(self.rawdata.Energy) / len(set(self.rawdata.Energy))

    @memoized_property
    def num_bins(self):
        return len(set(self.rawdata.Energy))

    def scan(self, N, scan_rng=0, unstacked=pd.DataFrame(np.array([]))):
        return self.rawdata.iloc[N*self.num_bins:(N+1)*self.num_bins, :].copy()

@register_xasclass
class BL11012Spectra(BL6311Spectra):
    colmap = {'TEY signal': 'Data', 'AI 3 Izero': 'Reference','Beamline Energy':'Energy'} # TODO: complexe normalization with external file
    skiprows = 15

    @staticmethod
    def validate(filename,skiprows=skiprows):
        BL6311Spectra.validate(filename,skiprows)


def tests():
    files = ['/home/rp/data/XAS/TrajScan21930.txt', #6311
             '/home/rp/data/XAS/TrajScan18764_0001.txt', #6311
             '/home/rp/data/XAS/DB_Cedge_000006.dat', #632
             '/home/rp/data/XAS/Fe_L_DB_29733.txt', #11012
             '/home/rp/data/XAS/O_K_DB_29726.txt'] #11012
    for s in files:
        print 'Opening:',s
        s = open(s)
        # print s.rawdata
        s.treat()
        # print s.scan(0)
        for scan in s.scans:
            print len(np.array(scan.Energy)),\
                len(np.array(scan.I_norm))



if __name__ == '__main__':
    tests()


from pyqtgraph.graphicsItems import GradientEditorItem
from collections import OrderedDict

GradientEditorItem.__dict__['Gradients'] = OrderedDict([
    ('thermal', {'ticks': [(0.3333, (185, 0, 0, 255)), (0.6666, (255, 220, 0, 255)), (1, (255, 255, 255, 255)), (0, (0, 0, 0, 255))], 'mode': 'rgb'}),
    ('flame', {'ticks': [(0.2, (7, 0, 220, 255)), (0.5, (236, 0, 134, 255)), (0.8, (246, 246, 0, 255)), (1.0, (255, 255, 255, 255)), (0.0, (0, 0, 0, 255))], 'mode': 'rgb'}),
    ('yellowy', {'ticks': [(0.0, (0, 0, 0, 255)), (0.2328863796753704, (32, 0, 129, 255)), (0.8362738179251941, (255, 255, 0, 255)), (0.5257586450247, (115, 15, 255, 255)), (1.0, (255, 255, 255, 255))], 'mode': 'rgb'} ),
    ('bipolar', {'ticks': [(0.0, (0, 255, 255, 255)), (1.0, (255, 255, 0, 255)), (0.5, (0, 0, 0, 255)), (0.25, (0, 0, 255, 255)), (0.75, (255, 0, 0, 255))], 'mode': 'rgb'}),
    ('viridis', {'ticks': [(0.0, (68, 1, 84, 255)), (0.25, (58, 82, 139, 255)), (0.5, (32, 144, 140, 255)), (0.75, (94, 201, 97, 255)), (1.0, (253, 231, 36, 255))], 'mode': 'rgb'}),
    ('inferno', {'ticks': [(0.0, (0, 0, 3, 255)), (0.25, (87, 15, 109, 255)), (0.5, (187, 55, 84, 255)), (0.75, (249, 142, 8, 255)), (1.0, (252, 254, 164, 255))], 'mode': 'rgb'}),
    ('plasma', {'ticks': [(0.0, (12, 7, 134, 255)), (0.25, (126, 3, 167, 255)), (0.5, (203, 71, 119, 255)), (0.75, (248, 149, 64, 255)), (1.0, (239, 248, 33, 255))], 'mode': 'rgb'}),
    ('magma', {'ticks': [(0.0, (0, 0, 3, 255)), (0.25, (80, 18, 123, 255)), (0.5, (182, 54, 121, 255)), (0.75, (251, 136, 97, 255)), (1.0, (251, 252, 191, 255))], 'mode': 'rgb'}),
    ('spectrum', {'ticks': [(1.0, (255, 0, 255, 255)), (0.0, (255, 0, 0, 255))], 'mode': 'hsv'}),
    ('cyclic', {'ticks': [(0.0, (255, 0, 4, 255)), (1.0, (255, 0, 0, 255))], 'mode': 'hsv'}),
    ('greyclip', {'ticks': [(0.0, (0, 0, 0, 255)), (0.99, (255, 255, 255, 255)), (1.0, (255, 0, 0, 255))], 'mode': 'rgb'}),
    ('grey', {'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (255, 255, 255, 255))], 'mode': 'rgb'}),
])

from pyqtgraph.exporters import ImageExporter
from pyqtgraph.Qt import QtGui, QtCore, QtSvg, USE_PYSIDE
from pyqtgraph import functions as fn
import numpy as np

class fixedexporter(ImageExporter):
    def export(self, fileName=None, toBytes=False, copy=False):
        if fileName is None and not toBytes and not copy:
            if USE_PYSIDE:
                filter = ["*." + str(f) for f in QtGui.QImageWriter.supportedImageFormats()]
            else:
                filter = ["*." + bytes(f).decode('utf-8') for f in QtGui.QImageWriter.supportedImageFormats()]
            preferred = ['*.png', '*.tif', '*.jpg']
            for p in preferred[::-1]:
                if p in filter:
                    filter.remove(p)
                    filter.insert(0, p)
            self.fileSaveDialog(filter=filter)
            return

        targetRect = QtCore.QRect(0, 0, self.params['width'], self.params['height'])
        sourceRect = self.getSourceRect()

        # self.png = QtGui.QImage(targetRect.size(), QtGui.QImage.Format_ARGB32)
        # self.png.fill(pyqtgraph.mkColor(self.params['background']))
        w, h = self.params['width'], self.params['height']
        if w == 0 or h == 0:
            raise Exception("Cannot export image with size=0 (requested export size is %dx%d)" % (w, h))
        bg = np.empty((int(self.params['width']), int(self.params['height']), 4), dtype=np.ubyte)
        color = self.params['background']
        bg[:, :, 0] = color.blue()
        bg[:, :, 1] = color.green()
        bg[:, :, 2] = color.red()
        bg[:, :, 3] = color.alpha()
        self.png = fn.makeQImage(bg, alpha=True)

        ## set resolution of image:
        origTargetRect = self.getTargetRect()
        resolutionScale = targetRect.width() / origTargetRect.width()
        # self.png.setDotsPerMeterX(self.png.dotsPerMeterX() * resolutionScale)
        # self.png.setDotsPerMeterY(self.png.dotsPerMeterY() * resolutionScale)

        painter = QtGui.QPainter(self.png)
        # dtr = painter.deviceTransform()
        try:
            self.setExportMode(True, {'antialias': self.params['antialias'], 'background': self.params['background'],
                                      'painter': painter, 'resolutionScale': resolutionScale})
            painter.setRenderHint(QtGui.QPainter.Antialiasing, self.params['antialias'])
            self.getScene().render(painter, QtCore.QRectF(targetRect), QtCore.QRectF(sourceRect))
        finally:
            self.setExportMode(False)
        painter.end()

        if copy:
            QtGui.QApplication.clipboard().setImage(self.png)
        elif toBytes:
            return self.png
        else:
            self.png.save(fileName)
fixedexporter.register()


# -*- coding: utf-8 -*-
"""
@author: lbluque
"""
import time
from PySide import QtGui, QtCore
from PySide.QtUiTools import QUiLoader


class LoginDialog(QtGui.QWidget):
    """
    Class for user credential input and login to NERSC/SPOT through NEWT/SPOT API

    Attributes
    ----------
    ui : QtGui.QWidget
        Widget specifying ui. Read from .ui file
    login_slot : Slot/function
        Slot to recieve login button clicked signal


    """
    REMOTE_NAMES = ('SPOT', 'NERSC', 'Globus')

    loginClicked = QtCore.Signal(tuple)
    logoutClicked = QtCore.Signal()
    sigLoggedIn = QtCore.Signal(bool)

    def __init__(self, remotename='SPOT', parent=None):
        super(LoginDialog, self).__init__(parent)

        if remotename not in self.REMOTE_NAMES:
            raise ValueError('Remote must be one of %s, not %s' % (self.REMOTE_NAMES, remotename))

        self.ui = QUiLoader().load('xicam/gui/login_widget.ui', self)
        self.ui.stackedWidget.setCurrentWidget(self.ui.login_page)
        self.ui.pass_box.setEchoMode(QtGui.QLineEdit.Password)

        self.ui.login_button.clicked.connect(self.handleLogin)
        self.ui.pass_box.returnPressed.connect(self.handleLogin)
        self.ui.host_box.returnPressed.connect(self.handleLogin)
        self.setStyleSheet('background-color:#111111;')

        self.hide()

        l = QtGui.QVBoxLayout()
        l.setContentsMargins(0, 0, 0, 0)
        l.addWidget(self.ui)
        self.setLayout(l)

        self._login_slot = None

    @property
    def login_slot(self):
        return self._login_slot

    @login_slot.setter
    def login_slot(self, slot):
        """
        Connect appropriate signals when setting slot
        """
        if self.login_slot is not None:
            self.loginClicked.disconnect(self.login_slot)
        self._login_slot = slot
        self.loginClicked.connect(self._login_slot)

    @QtCore.Slot(QtCore.Signal, bool)
    def loginRequest(self, login_clicked_slot, show_host=False):
        """
        Slot to receive login request signal and setup/show the loginDialog accordingly

        Parameters
        ----------
        login_clicked_slot : Slot
            Slot to set as the login_slot to call. This should be the clients constructor or login method
        show_host : bool
            Boolean to show the host QTextEdit to input a hostname
        """

        if show_host:
            self.ui.host_box.show()
        else:
            self.ui.host_box.hide()
        self.ui.user_box.setFocus()
        self.login_slot = login_clicked_slot
        self.setCurrentWidget(self.ui.login_page)
        self.show()

    def handleLogin(self):
        """
        Handles the login button clicked signal and calls the login_slot if all input fields are satisfied
        """
        host, usr, pwd = self.ui.host_box.text(), self.ui.user_box.text(), self.ui.pass_box.text()
        if usr == '':
            QtGui.QMessageBox.warning(self, 'Username missing', 'You forgot to mention who you are!')
        elif pwd == '':
            QtGui.QMessageBox.warning(self, 'Password missing',
                                      'You need to provide proof that you really are {}!'.format(usr))
        elif host == '' and not self.ui.host_box.isHidden():
            QtGui.QMessageBox.warning(self, 'Host missing',
                                      'Xi-cam can not guess what host you want to connect to yet!')
        else:
            self.setCurrentWidget(self.ui.progress_page)
            self.progressMessage('Logging in...')
            self.startProgress()
            credentials = {'username': usr, 'password': pwd} if self.ui.host_box.isHidden() \
                else {'username': usr, 'password': pwd, 'host': host}
            self.loginClicked.emit(credentials)

    def loginResult(self, status):
        """
        Slot called upon finishing the login_slot call. Emits sigLoggedIn with the status

        Parameters
        ----------
        status : bool
            Status of login. True if successful
        """
        self.stopProgress()
        self.setCurrentWidget(self.ui.login_page)
        if status is True:
            for box in (self.ui.pass_box, self.ui.user_box, self.ui.host_box):
                box.clear()
            self.hide()
        else:
            self.ui.pass_box.clear()
        self.sigLoggedIn.emit(status)


    def setCurrentWidget(self, widget):
        """
        Used to set the page in the loginDialogs stackedWidget
        """
        self.ui.stackedWidget.setCurrentWidget(widget)

    def startProgress(self):
        """
        Starts the progress bar pulsation
        """
        self.ui.progressBar.setRange(0, 0)

    def stopProgress(self):
        """
        Stops the progress bar pulsation
        """
        self.ui.progressBar.setRange(0, 1)

    def progressMessage(self, message):
        # Not working for some reason
        self.ui.progress_label.setText(message)



#! /usr/bin/env python

# from image_load import *
# import loader
import numpy as np
from pyFAI import geometry

from PySide import QtGui
import msg


try:
    from cWarpImage import warp_image
except ImportError:
    msg.logMessage("Remeshing C extension is NOT LOADED!",msg.ERROR)


def calc_q_range(lims, geometry, alphai, cen):
    nanometer = 1.0E+09
    sdd = geometry.get_dist() * nanometer
    wavelen = geometry.get_wavelength() * nanometer
    pixel1 = geometry.get_pixel1() * nanometer
    pixel2 = geometry.get_pixel2() * nanometer

    # calculate q-range for the image
    y = np.array([0, lims[1] - 1], dtype=np.float) * pixel1
    z = np.array([0, lims[0] - 1], dtype=np.float) * pixel2
    y, z = np.meshgrid(y, z)
    y -= cen[0]
    z -= cen[1]

    # calculate angles 
    tmp = np.sqrt(y ** 2 + sdd ** 2)
    cos2theta = sdd / tmp
    sin2theta = y / tmp
    tmp = np.sqrt(z ** 2 + y ** 2 + sdd ** 2)
    cosalpha = sdd / tmp
    sinalpha = z / tmp
    k0 = 2. * np.pi / wavelen

    # calculate q-values of each corner
    qx = k0 * (cosalpha * cos2theta - np.cos(alphai))
    qy = k0 * cosalpha * sin2theta
    qz = k0 * (sinalpha + np.sin(alphai))
    qp = np.sign(qy) * np.sqrt(qx ** 2 + qy ** 2)
    q_range = [qp.min(), qp.max(), qz.min(), qz.max()]
    return q_range, k0


def remesh(image, filename, geometry, alphai):
    shape = image.shape
    center = np.zeros(2, dtype=np.float)
    pixel = np.zeros(2, dtype=np.float)

    # get calibrated parameters
    nanometer = 1.0E+09
    sdd = geometry.get_dist() * nanometer
    pixel[0] = geometry.get_pixel1() * nanometer
    pixel[1] = geometry.get_pixel2() * nanometer
    center[0] = geometry.get_poni2() * nanometer
    center[1] = shape[0] * pixel[0] - geometry.get_poni1() * nanometer

    # calculate q values
    qrange, k0 = calc_q_range(image.shape, geometry, alphai, center)

    # uniformly spaced q-values for remeshed image
    nqz = image.shape[0]
    dqz = (qrange[3] - qrange[2]) / (nqz - 1)
    nqp = np.int((qrange[1] - qrange[0]) / dqz)
    qvrt = np.linspace(qrange[2], qrange[3], nqz)
    qpar = qrange[0] + np.arange(nqp) * dqz
    qpar, qvrt = np.meshgrid(qpar, qvrt)

    try:
        qimg = warp_image(image, qpar, qvrt, pixel, center, alphai, k0, sdd, 0)
        return np.rot90(qimg, 3), np.rot90(qpar, 3), np.rot90(qvrt, 3)
    except Exception:

        # find inverse map
        cosi = np.cos(alphai)
        sini = np.sin(alphai)

        sina = (qvrt / k0) - sini
        cosa2 = 1 - sina ** 2
        cosa = np.sqrt(cosa2)
        tana = sina / cosa

        t1 = cosa2 + cosi ** 2 - (qpar / k0) ** 2
        t2 = 2. * cosa * cosi
        cost = t1 / t2
        cost[t1 > t2] = 0
        with np.errstate(divide='ignore'):
            tant = np.sign(qpar) * np.sqrt(1. - cost ** 2) / cost
            tant[cost == 0] = 0

        # F : (qp,qz) --> (x,y)
        map_x = (tant * sdd + center[0]) / pixel[0]
        cost[cost < 0] = 1
        with np.errstate(divide='ignore'):
            map_y = (tana * sdd / cost + center[1]) / pixel[1]


        # compute null space
        nrow, ncol = image.shape
        m1 = t1 > t2
        m2 = np.logical_or(map_x < 0, map_x > ncol - 1)
        m3 = np.logical_or(map_y < 0, map_y > nrow - 1)

        mask = np.logical_or(np.logical_or(m1, m2), m3)
        map_x[mask] = -1
        map_y[mask] = -1

        qsize = map_x.size
        qshape = map_x.shape
        qimg = np.fromiter((image[i, j] for i, j in np.nditer([map_y.astype(int),
                                                               map_x.astype(int)])), dtype=np.float,
                           count=qsize).reshape(qshape)
        qimg[mask] = 0.
        return np.rot90(qimg, 3), np.rot90(qpar, 3), np.rot90(qvrt, 3)

def remeshqarray(image, filename, geometry, alphai):
    shape = image.shape
    center = np.zeros(2, dtype=np.float)
    pixel = np.zeros(2, dtype=np.float)

    # get calibrated parameters
    nanometer = 1.0E+09
    pixel[0] = geometry.get_pixel1() * nanometer
    pixel[1] = geometry.get_pixel2() * nanometer
    center[0] = geometry.get_poni2() * nanometer
    center[1] = shape[0] * pixel[0] - geometry.get_poni1() * nanometer

    # calculate q values
    qrange, k0 = calc_q_range(image.shape, geometry, alphai, center)

    # uniformly spaced q-values for remeshed image
    nqz = image.shape[0]
    dqz = (qrange[3] - qrange[2]) / (nqz - 1)
    nqp = np.int((qrange[1] - qrange[0]) / dqz)
    qvrt = np.linspace(qrange[2], qrange[3], nqz)
    qpar = qrange[0] + np.arange(nqp) * dqz
    qpar, qvrt = np.meshgrid(qpar, qvrt)

    return np.rot90(qpar, 3), np.rot90(qvrt, 3)


if __name__ == "__main__":
    # create a test case with known geometry
    import fabio
    import pylab as plt
    import time
    import pdb

    filename = '/Users/dkumar/Data/examples/Burst/calibration/AGB_5S_USE_2_2m.edf'
    image = fabio.open(filename).data
    sdd = 0.283351
    cen_y = 0.005363
    cen_x = 0.040123
    pixel = 0.000172
    geo = geometry.Geometry(sdd, cen_y, cen_x, 0, 0, 0)
    geo.set_wavelength(1.23984E-10)
    geo.set_pixel1(0.172E-03)
    geo.set_pixel2(0.172E-03)
    t0 = time.clock()
    alphai = np.deg2rad(0.14)
    qimg, qpar, qvrt = remesh(image, filename, geo, alphai)
    qimg.tofile("img.bin")
    qpar.tofile("qpar.bin")
    qvrt.tofile("qvrt.bin")
    t1 = time.clock() - t0
    msg.logMessage("remesh clock time = %f" % t1,msg.DEBUG)
    plt.imshow(np.log(qimg + 5), cmap=plt.cm.autumn_r, interpolation='Nearest',
               extent=[qpar.min(), qpar.max(), -1 * qvrt.max(), -1 * qvrt.min()])
    plt.show()




# REF: http://stackoverflow.com/questions/23728401/pyside-crashing-python-when-emitting-none-between-threads

_PYSIDE_NONE_SENTINEL = object()


def pyside_none_wrap(var):
    """None -> sentinel. Wrap this around out-of-thread emitting."""
    if var is None:
        return _PYSIDE_NONE_SENTINEL
    return var


def pyside_none_deco(func):
    """sentinel -> None. Decorate callbacks that react to out-of-thread
    signal emitting.

    Modifies the function such that any sentinels passed in
    are transformed into None.
    """

    def sentinel_guard(arg):
        if arg is _PYSIDE_NONE_SENTINEL:
            return None
        return arg

    def inner(*args, **kwargs):
        newargs = map(sentinel_guard, args)
        newkwargs = {k: sentinel_guard(v) for k, v in kwargs.iteritems()}
        return func(*newargs, **newkwargs)

    return inner


import numpy as np  # Import important packages
from pyqtgraph.Qt import QtCore, QtGui
import pyqtgraph as pg
import glob
from PIL import Image
from collections import OrderedDict
import os
import re


def calcscale(imv):  # Defines calcscale function
    """

    """
    image = imv.getProcessedImage()

    scale = imv.scalemax / float(image[imv.currentIndex].shape[1])
    return scale


class imagetimeline(list):  # Sets up the image so it will fin the the viewer

    @property
    def shape(self):  # Defines shape function
        return (len(self), self[-1].shape[0], self[-1].shape[0])

    def __getitem__(self, item):  # Defines getitem function
        return list.__getitem__(self, item)

    @property
    def ndim(self):  # Defines ndim functionq
        return 3

    @property
    def size(self):  # Defines size functon
        return sum(map(np.size, self))

    @property
    def max(self):
        return max(map(np.max, self))

    @property
    def min(self):
        return min(map(np.min, self))

    @property
    def dtype(self):
        return type(self[0][0, 0])



class TimelineView(pg.ImageView):  # Beginnings the class Timelineview

    sigImageChanged = QtCore.Signal()

    def __init__(self, scalemax, *args, **kwargs):
        super(TimelineView, self).__init__(*args, **kwargs)
        self.scalemax = scalemax

        self.view_label = QtGui.QLabel(self)
        self.view_label.setText('No: ')
        self.view_number = QtGui.QSpinBox(self)
        self.view_number.setReadOnly(True)
        self.view_number.setMaximum(10000)
        self.ui.gridLayout.addWidget(self.view_label, 1, 1, 1, 1)
        self.ui.gridLayout.addWidget(self.view_number, 1, 2, 1, 1)

        self.label = QtGui.QLabel(parent=self)
        self.ui.gridLayout.addWidget(self.label, 1, 0, 1, 1)



    def quickMinMax(self, data):  # Defines quickMinMax functon
        return min(map(np.min, data)), max(map(np.max, data))

    def updateImage(self, autoHistogramRange=True):  # Defines updateImage functon
        if self.image is None:
            return

        scale = calcscale(self)  # Scales the image
        image = self.getProcessedImage()

        if autoHistogramRange:  # Sets the Y axis intensity bar
            self.ui.histogram.setHistogramRange(self.levelMin, self.levelMax)
        if self.axes['t'] is None:
            self.imageItem.updateImage(image)
        else:
            self.ui.roiPlot.show()
            self.imageItem.updateImage(image[self.currentIndex])

        self.imageItem.resetTransform()  # Resets the scale up below
        self.imageItem.scale(scale, scale)  # Scales up by the factor of scale
        print 'Image shape' + str(image.shape)
        print 'Scale set to: ' + str(scale)

        self.view_number.setValue(self.currentIndex)
        self.sigImageChanged.emit()




class fftView(QtGui.QTabWidget):
    def __init__(self, *args, **kwargs):
        super(fftView, self).__init__(*args, **kwargs)
        self.img_dict = OrderedDict()

    def add_images(self, image_list, do_fft=True, loadingfactors=None):

        self.clear()

        if not image_list:
            return
        if type(image_list) == dict:
            for path, img in image_list.iteritems():
                try:
                    img = np.array(img)
                    len(img)
                    flag = True
                    if do_fft: img = self.do_fft(img)

                    for item in self.img_dict.itervalues():
                        if np.array_equal(img, item): flag = False
                    if flag:
                        self.img_dict[path.split('/')[-1]] = img
                except TypeError:
                    continue
        else:
            for img in image_list:
                try:
                    img = np.array(img)
                    len(img)
                    flag = True
                    if do_fft: img = self.do_fft(img)

                    for item in self.img_dict.itervalues():
                        if np.array_equal(img, item): flag = False
                    if flag:
                        self.img_dict[len(self.img_dict)] = img
                except TypeError:
                    continue

        # sort img dictionary by key
        img_dict = OrderedDict()
        for key in sorted(self.img_dict.keys()):
            img_dict[key] = self.img_dict[key]
        del self.img_dict
        self.img_dict = img_dict

        data = imagetimeline(self.img_dict.itervalues())
        sizemax = max(map(np.shape, data))[0]

        view = TimelineView(sizemax)
        view.setImage(data)

        scale = calcscale(view)  # Sets up the scale
        view.imageItem.resetTransform()
        view.imageItem.scale(scale, scale)
        view.autoRange()
        view.getHistogramWidget().setHidden(False)
        view.ui.roiBtn.setHidden(True)
        view.ui.menuBtn.setHidden(True)
        view.sigImageChanged.connect(self.imageNameChanged)

        if loadingfactors is None:
            self.addTab(view, u"Tile " + str(1))
        else:
            self.addTab(view, str(loadingfactors))
        self.tabBar().hide()

        view.sigImageChanged.emit()

    def do_fft(self, img):
        """
        Returns absolute value of 2-D FFT of image. Assumes 2-D image
        """

        return np.log(np.fft.fftshift(np.abs(np.fft.fft2(img)**2)) + 1e-10)


    def open_from_rmcView(self, image_list):
        images = {}
        for lst in image_list:
            path = "/"
            for item in lst:
                path = os.path.join(path, item)
            img = Image.open(path).convert('L')
            img = np.array(img)
            images[path] = img

        self.add_images(images)

    def imageNameChanged(self):
        view = self.currentWidget()
        for key, val in self.img_dict.iteritems():
            if np.array_equal(val, view.image[view.currentIndex]):
                self.currentWidget().label.setText(key)


class rmcView(QtGui.QTabWidget):

    def __init__(self, root, loadingfactors=[None]):
        super(rmcView, self).__init__()

        self.image_list = []
        self.image_dict = OrderedDict()

        paths = glob.glob(os.path.join(root,
                                       '[0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9]_model.tif'))

        indices = dict(zip(paths, [re.findall('\d{4}', os.path.basename(path)) for path in paths]))

        tiles = dict()

        for path, ind in indices.iteritems():
            if int(ind[1]) in tiles:
                tiles[int(ind[1])].append(path)
            else:
                tiles[int(ind[1])] = [path]
            self.image_list.append(path.split('/'))

        for tile, loadingfactor in zip(tiles, loadingfactors):
            images = []
            paths = sorted(tiles[tile])
            for path in paths:
                img = Image.open(path).convert('L')
                img = np.array(img)

                print path  # Prints the path
                print img.shape  # Prints the shape of the array

                images.append(img)
                self.image_dict[path.split('/')[-1]] = img

            data = imagetimeline(images)

            sizemax = max(map(np.shape, data))[0]

            view = TimelineView(sizemax)
            view.setImage(data)

            scale = calcscale(view)  # Sets up the scale
            view.imageItem.resetTransform()
            view.imageItem.scale(scale, scale)
            view.autoRange()
            view.getHistogramWidget().setHidden(False)
            view.ui.roiBtn.setHidden(True)
            view.ui.menuBtn.setHidden(True)
            view.sigImageChanged.connect(self.imageNameChanged)

            if loadingfactors is None:
                self.addTab(view, u"Tile " + str(tile + 1))
            else:
                self.addTab(view, str(loadingfactor))

            view.sigImageChanged.emit()




    def addNewImages(self, root, loadingfactors=[None]):

        self.clear()

        paths = glob.glob(os.path.join(root,
                                       '[0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9]_[0-9][0-9][0-9][0-9]_model.tif'))

        indices = dict(zip(paths, [re.findall('\d{4}', os.path.basename(path)) for path in paths]))

        tiles = dict()


        for path, ind in indices.iteritems():
            if int(ind[1]) in tiles:
                tiles[int(ind[1])].append(path)
            else:
                tiles[int(ind[1])] = [path]

            if path.split('/') not in self.image_list:
                self.image_list.append(path.split('/'))

        for tile, loadingfactor in zip(tiles, loadingfactors):
            images = []
            paths = sorted(tiles[tile])
            for path in paths:
                img = Image.open(path).convert('L')
                img = np.array(img)

                print path  # Prints the path
                print img.shape  # Prints the shape of the array

                images.append(img)
                self.image_dict[path.split('/')[-1]] = img

            data = imagetimeline(images)
            sizemax = max(map(np.shape, data))[0]

            view = TimelineView(sizemax)
            view.setImage(data)

            scale = calcscale(view)  # Sets up the scale
            view.imageItem.resetTransform()
            view.imageItem.scale(scale, scale)
            view.autoRange()
            view.getHistogramWidget().setHidden(False)
            view.ui.roiBtn.setHidden(True)
            view.ui.menuBtn.setHidden(True)
            view.sigImageChanged.connect(self.imageNameChanged)


            if loadingfactors is None:
                self.addTab(view, u"Tile " + str(tile + 1))
            else:
                self.addTab(view, str(loadingfactor))

            view.sigImageChanged.emit()


    def imageNameChanged(self):

        view = self.currentWidget()
        for key, val in self.image_dict.iteritems():
            if np.array_equal(val, view.image[view.currentIndex]):
                self.currentWidget().label.setText(key)




if __name__ == '__main__':  # Start Qt event loop unless running in interactive mode.
    import sys

    app = QtGui.QApplication([])  # Launches an app
#    root = '/Users/austinblair/Downloads/test_20150714_144045/'
    root = '/home/hparks/Desktop/processed_20161201_170824'

#    win = QtGui.QMainWindow()  # Create window with two ImageView widgets
#    win.resize(800, 800)

    win = QtGui.QStackedWidget()
    win.setWindowTitle('pyqtgraph example: Hiprmc ')

#    win.setCentralWidget(rmcView(root))
    win.addWidget(rmcView(root))

    win.show()

    if (sys.flags.interactive != 1) or not hasattr(QtCore, 'PYQT_VERSION'):
        QtGui.QApplication.instance().exec_()



#! /usr/bin/env python
# --coding: utf-8 --
import itertools
import numpy as np
from numpy.linalg import norm
from scipy.optimize import root
import sgexclusions
from xicam import config

def volume(a, b, c, alpha=None, beta=None, gamma=None):
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray) and \
            isinstance(c, np.ndarray):
        return abs(np.dot(a, np.cross(b, c)))
    elif isinstance(a, float) and isinstance(b, float) and isinstance(c, float):
        if alpha is None:
            alpha = np.pi / 2
        if beta is None:
            beta = np.pi / 2
        if gamma is None:
            gamma = np.pi / 2
        abc2 = (a * b * c ) ** 2
        t1 = np.cos(alpha) * np.cos(beta) * np.cos(gamma)
        t2 = np.cos(alpha) ** 2 + np.cos(beta) ** 2 + np.cos(gamma) ** 2
        return np.sqrt(abc2 * (1 - t2 + 2 * t1))
    else:
        raise ValueError('vectors must be either numpy.ndarray or real types')


def crystal2sample(a, b, c, alpha=None, beta=None, gamma=None):
    R = np.zeros((3, 3), dtype=np.float)

    # don't need to do all this if a,b,c are vectors
    # we should convert them to vectors if they are not
    if alpha is None:
        alpha = np.pi / 2
    if beta is None:
        beta = np.pi / 2
    if gamma is None:
        gamma = np.pi / 2
    t1 = np.cos(gamma) * np.cos(beta) - np.cos(alpha)
    vol = volume(a, b, c, alpha, beta, gamma)
    R[0, 0] = a
    R[0, 1] = b * np.cos(gamma)
    R[0, 2] = c * np.cos(beta)
    R[1, 1] = b * np.sin(gamma)
    R[1, 2] = -c * t1 / np.sin(gamma)
    R[2, 2] = vol / (a * b * np.sin(gamma))
    return R


def orientation_uvw(uvw, R):
    e = R.dot(uvw)
    return e / norm(e)


def orientation_hkl(hkl, R):
    h = hkl[0]
    k = hkl[1]
    l = hkl[2]
    if h == 0 and k == 0 and l == 0:
        raise ValueError("one of [h k l] must be non-zero")

    A_c = np.zeros(3, dtype=np.float)
    B_c = np.zeros(3, dtype=np.float)
    C_c = np.zeros(3, dtype=np.float)

    # vector Ac
    if not h == 0:
        A_c[0] = 1. / hkl[0]
    else:
        A_c[0] = 1.;
        if not k == 0:
            A_c[1] = 1
        else:
            A_c[2] = 1

    # vector Bc
    if not k == 0:
        B_c[1] = 1. / hkl[1]
    else:
        B_c[1] = 1.;
        if not l == 0:
            B_c[2] = 1
        else:
            B_c[0] = 1

    # vector Cc
    if not l == 0:
        C_c[2] = 1. / hkl[2]
    else:
        C_c[2] = 1.
        if not h == 0:
            C_c[0] = 1.;
        else:
            C_c[1] = 1.;
    AsBs = np.inner(R, (B_c - A_c))
    BsCs = np.inner(R, (C_c - B_c))
    t1 = np.cross(AsBs, BsCs)
    return t1 / norm(t1)


def RotationMatrix(u, theta):
    R = np.zeros((3, 3), dtype=np.float)
    cost = np.cos(theta)
    sint = np.sin(theta)
    R[0, 0] = cost + u[0] ** 2 * ( 1 - cost )
    R[0, 1] = u[0] * u[1] * ( 1 - cost ) - u[2] * sint
    R[0, 2] = u[0] * u[2] * ( 1 - cost ) + u[1] * sint
    R[1, 0] = u[0] * u[1] * ( 1 - cost ) - u[2] * sint
    R[1, 1] = cost + u[1] ** 2 * ( 1 - cost )
    R[1, 2] = u[1] * u[2] * ( 1 - cost ) - u[0] * sint
    R[2, 0] = u[0] * u[2] * ( 1 - cost ) - u[1] * sint
    R[2, 1] = u[1] * u[2] * ( 1 - cost ) + u[0] * sint
    R[2, 2] = cost + u[2] ** 2 * ( 1 - cost )
    return R;


def reciprocalvectors(a, b, c):
    V = volume(a, b, c)
    mra = 2 * np.pi / V * np.cross(b, c)
    mrb = 2 * np.pi / V * np.cross(c, a)
    mrc = 2 * np.pi / V * np.cross(a, b)
    vecs = np.vstack([mra, mrb, mrc])
    return vecs


def equations(x, G, alphai, k):
    ci = np.cos(alphai)
    si = np.sin(alphai)
    y = np.zeros(4, dtype=float)
    y[0] = x[0] ** 2 + x[2] ** 2 - G[0] ** 2 - G[1] ** 2
    y[1] = x[0]  - x[1] * ci - x[3] * si
    y[2] = x[1] * si - x[3] * ci + G[2]
    y[3] = (x[1] - k) ** 2 + x[2] ** 2 + x[3] ** 2 - k ** 2
    return y

def jacobian(x, G, alphai, k):
        ci = np.cos(alphai)
        si = np.sin(alphai)
        jac = np.zeros((4,4),dtype=float)
        jac[0,0] = 2 * x[0]
        jac[0,1] = 1
        jac[1,1] = -ci
        jac[1,2] = si
        jac[1,3] = 2 * (x[1] - k)
        jac[2, 0] = 2 * x[2]
        jac[2,3] = 2 * x[2]
        jac[3,1] = -si
        jac[3,2] = -ci
        jac[3,3] = 2 * x[3]
        return jac

def reflection_condtion(hkl, unitcell, space_grp):
    if unitcell is None:
        return True
    tot = 0.
    c = -2 * np.pi * np.complex(0, 1)
    for pos in unitcell:
        tot += np.real(np.exp(c * np.dot(pos, hkl)))
    if abs(np.real(tot)) < 1.0E-06:
        return False
    else:
        return True


def alpha_exit(qz, alphai, k):
    alf = np.arcsin(qz / k - np.sin(alphai))
    return alf

def theta_exit(q, alpha, alphai, k):
    t1 = np.cos(alpha)**2 + np.cos(alphai)**2 - (q[0]**2 + q[1]**2)/(k*k)
    t2 = 2 * np.cos(alpha) * np.cos(alphai)
    return np.sign(q[1]) * np.arccos(t1/t2)

def angles_to_pixels(angles, center, sdd, pixel_size=None):
    if pixel_size is None:
        pixel_size = [172e-6, 172e-6]
    if np.NaN in angles:
        return np.NAN, np.NAN
    tan_2t = np.tan(angles[:, 0])
    tan_al = np.tan(angles[:, 1])
    x = tan_2t * sdd
    px = sdd * tan_2t / pixel_size[0] #+ center[0]
    py = np.sqrt(sdd ** 2 + x ** 2) * tan_al  / pixel_size[1] #+ center[1]
    pixels = np.zeros((px.size, 2))
    pixels[:,0] = px
    pixels[:,1] = py
    return pixels.astype(int)

class peak(object):
    def __init__(self, mode):
        self.mode = mode # either 'Transmission' or 'Reflection'
        self.hkl = None
        self.x = None
        self.y = None
        self.twotheta = None
        self.alphaf = None
        self.qpar = None
        self.qz = None

    def pos(self):
        return self.x, self.y

    # qz, alphaf and y-poition are going to be different
    def copy(self):
        val = peak(self.mode)
        val.twotheta = self.twotheta
        val.qpar = self.qpar
        val.hkl = self.hkl
        val.x = self.x
        return val

    def position(self, center, sdd, pixels):
        tan_2t = np.tan(self.twotheta)
        tan_al = np.tan(self.alphaf)
        x= tan_2t * sdd
        self.x = sdd * tan_2t / pixels# + config.activeExperiment.center[0]
        self.y = np.sqrt(sdd ** 2 + x ** 2) * tan_al / pixels# + config.activeExperiment.center[1]

    def isAt(self, pos):
        if np.isnan(self.twotheta): return False
        return int(self.x) == int(pos.x()) and int(self.y) == int(pos.y())

    def __str__(self):
        s = u"Peak type: {}\n".format(self.mode)
        s += u"Lattice vector (h,k,l): {}\n".format(self.hkl)
        if self.twotheta is not None: s += u"2\u03B8: {}\n".format(self.twotheta)
        if self.alphaf is not None: s += u"\u03B1f: {}\n".format(self.alphaf)
        if self.qpar is not None: s += u"q\u2225: {}\n".format(self.qpar/1e10)
        if self.qz is not None: s += u"q\u27c2: {}\n".format(self.qz/1e10)
        return s

def qvalues(twotheta, alphaf, alphai, wavelen):
    k = 2 * np.pi / wavelen
    qx = k * (np.cos(alphaf) * np.cos(twotheta) - np.cos(alphai))
    qy = k * np.cos(alphaf) * np.sin(twotheta)
    qz = k * np.sin(alphaf) + np.sin(alphai)
    return np.sqrt(qx**2 + qy**2), qz

def dwba_componets(Gz, k, nu, alphai):
    alphaf = np.zeros(4, dtype=np.float)
    gprime = Gz/k
    t1 = nu**2 - np.cos(alphai)**2
    if np.real(t1) < 0:
        print 'Error: incoming angle is below the critical angle'
        return None
    cprime = np.sqrt(t1)
    cosa1 = nu**2 - (gprime + cprime)**2 
    cosa2 = nu**2 - (gprime - cprime)**2 
    a1 = np.arccos(np.sqrt(cosa1))
    a2 = np.arccos(np.sqrt(cosa2))
    #TODO: this is a hack a1 should always be transmission
    if abs(a2) > abs(a1):
        alphaf[0] = a1
        alphaf[1] = a2
    else:
        alphaf[0] = a2
        alphaf[1] = a1
    
    return alphaf

def find_peaks(a, b, c, alpha=None, beta=None, gamma=None, normal=None,
               norm_type="uvw", wavelen=0.123984e-9, refdelta=2.236E-06, refbeta=-1.8790E-09, order=5, unitcell=None, space_grp=None):
    # rotation matrix from crystal coordinates for sample coordinates
    if alpha is not None: alpha = np.deg2rad(alpha)
    if beta is not None: beta = np.deg2rad(beta)
    if gamma is not None: gamma = np.deg2rad(gamma)

    A = crystal2sample(a, b, c, alpha, beta, gamma)
    # unit vector normal to sample plane
    normal = np.array(normal)
    if norm_type == "xyz":
        e_norm = normal / norm(normal)
    elif norm_type == "hkl":
        if not normal.dtype.type is np.int_:
            raise TypeError("hkl type must be integer datatype")
        e_norm = orientation_hkl(normal, A)
    elif norm_type == "uvw":
        n1 = np.array(normal, dtype=float)
        e_norm = orientation_uvw(n1, A)
    else:
        raise ValueError("error: unable to process normal direction")

    # Rotation axis
    e_z = np.array([0, 0, 1], dtype=float)
    e_r = np.cross(e_norm, e_z)
    e_r /= (norm(e_r) if norm(e_r) != 0 else 1)

    # Angle of rotation
    theta = np.arccos(np.dot(e_z, e_norm))
    # Rotation Matrix
    R = RotationMatrix(e_r, theta)

    # Lattice vectors in sample frame
    V = np.dot(R, A)
    a = V[:, 0]
    b = V[:, 1]
    c = V[:, 2]
    RV = reciprocalvectors(a, b, c)

    nu = 1 - np.complex(refdelta, refbeta)
    HKL = itertools.product(range(-order, order + 1), repeat=3)
    alphai = np.deg2rad(config.activeExperiment.getvalue('Incidence Angle (GIXS)'))
    k = 2 * np.pi / wavelen
    peaks = list()
    for hkl in HKL:
        if not sgexclusions.check(hkl, space_grp): continue
        if (reflection_condtion(hkl, unitcell, space_grp)):
            G = RV[0, :] * hkl[0] + RV[1, :] * hkl[1] + RV[2, :] * hkl[2]
            qpar = np.sqrt(G[0]**2 + G[1]**2)
            alpha = alpha_exit(G[2], alphai, k)
            theta = theta_exit(G, alpha, alphai, k)

            # compute DWBA components
            alphaf = dwba_componets(G[2], k, nu, alphai)
            if alphaf is None:
                continue

            # get the transmission peaks
            transmission = peak('Transmission')
            transmission.hkl = hkl
            transmission.twotheta = theta
            transmission.alphaf = alphaf[0]
            transmission.qpar = qpar
            transmission.qz = G[2]
            peaks.append(transmission)

            # calulate reflection peaks
            reflection = peak('Reflection')
            reflection.hkl = hkl
            reflection.twotheta = theta
            reflection.alphaf = alphaf[1]
            reflection.qpar = qpar
            reflection.qz = G[2]
            peaks.append(reflection)

    return peaks


if __name__ == '__main__':
    ang1 = np.deg2rad(90.)
    ang2 = np.deg2rad(90.)
    ang3 = np.deg2rad(90.)
    n = np.array([0, 0, 1], dtype=float)
    peaks = find_peaks(10.E-09, 10.E-09, 10.E-09, alpha=ang1, beta=ang2, gamma=ang3, normal=n, norm_type='uvw', wavelen=0.123984E-09,
                       order=2)
    trans_peaks = []
    refle_peaks = []
    for key in peaks:
        trans_peaks.append(peaks[key][0])
        refle_peaks.append(peaks[key][1])

    center = np.array([672,224], dtype=int)
    tr = np.array(trans_peaks, dtype=float)
    p1 = angles_to_pixels(tr,center, sdd=1.84)
    rf = np.array(refle_peaks, dtype=float)
    p2 = angles_to_pixels(rf,center, sdd=1.84)
    print str(p1)
    print str(p2)



import os
import time
import process
from PySide import QtCore
import multiprocessing
from xicam import debugtools


class daemon(QtCore.QThread):
    """
    This starts a daemon running as a QThread which watches a directory for new files and sends them to process.py
    Processing is done in separate Processes, splitting tasks between each core. Files that already exist can also
    be processed.
    """


    num_cores = multiprocessing.cpu_count()
    def __init__(self, path, experiment, procold=False):
        super(daemon, self).__init__()
        self.procold = procold
        self.experiment = experiment
        self.path = path
        self.exiting = False
        self.childfiles = set(os.listdir(path))


    def run(self):

        if self.procold:
            self.processfiles(self.path, self.childfiles)

        try:
            while not self.exiting:
                time.sleep(.1)
                self.checkdirectory()  # Force update; should not have to do this -.-
        except KeyboardInterrupt:
            pass

    def stop(self):
        self.exiting = True
        print ("thread stop - %s" % self.exiting)

    def __del__(self):
        self.exiting = True
        self.wait()

    @debugtools.timeit
    def processfiles(self, path, files):
        """
        distribute new files to cores for processing. Ignores .nxs.
        """

        files = [f for f in files if not os.path.splitext(f)[1] == '.hdf']
        if files:
            print os.path.splitext(path)[1]

            jobs = []
            p = None
            # filter paths
            files = list(chunks(files, self.num_cores))

            for i in range(self.num_cores):
                p = multiprocessing.Process(target=process.process, args=(path, files[i], self.experiment))
                jobs.append(p)
                p.start()

            while p.is_alive():  # TODO: change to wait for ALL threads?
                time.sleep(.1)



    def checkdirectory(self):
        """
        Checks a directory for new files, comparing what files are there now vs. before
        """
        updatedchildren = set(os.listdir(self.path))
        newchildren = updatedchildren - self.childfiles
        self.childfiles = updatedchildren
        self.processfiles(self.path, list(newchildren))



def chunks(l, n):
    """
    Yield successive n chunks from l.
    """
    chunksize = int(len(l) / n)
    for i in xrange(0, n, 1):
        yield l[i * chunksize:(i + 1) * chunksize]


if __name__ == '__main__':
    path = '/Users/rp/YL1031'
    # files = os.listdir(path)
    experiment = None
    procold = True
    d = daemon(path, experiment, procold)
    d.run()



import numpy as np
import tomopy
import afnumpy as afnp

def padmat(x,siz,value):
# function y=padmat(x,size,vals)
# pads x to size with constant values (vals), 
# centers the matrix in the middle.
    n=siz[0]
    if siz.size < 2:
      m=n
    elif siz.size == 2:
      m=siz[1]
    else:
      (n,m) = siz.shape
    
    [N,M]=x.shape

    y=np.zeros((n,m))+value
    y[0:N,0:M]=x
    y=np.roll(np.roll(y,np.int16(np.fix((n-N)/2)),axis=0),np.int16(np.fix((m-M)/2)),axis=1)
    return y


def padmat_v2(x,siz,value,y):
# function y=padmat(x,size,vals)
# pads x to size with constant values (vals), 
# centers the matrix in the middle.
    n=siz[0]
    if siz.size < 2:
      m=n
    elif siz.size == 2:
      m=siz[1]
    else:
      (n,m) = siz.shape
    
    [N,M]=x.shape
    y[0:N,0:M]=x
    y=np.roll(np.roll(y,afnp.int16(np.fix((n-N)/2)),axis=0),np.int16(afnp.fix((m-M)/2)),axis=1)
    return y



# --coding: utf-8 --
import pickle
import pyFAI
from pyFAI import geometry
from PySide import QtGui
from pyqtgraph.parametertree import Parameter
from pyqtgraph.parametertree import ParameterTree
from pyqtgraph.parametertree import parameterTypes as ptypes
import numpy as np
import yaml
from pipeline import pathtools
import os
from pipeline import msg
from pipeline import detectors


class settingstracker(ptypes.GroupParameter):
    settingspath = os.path.join(pathtools.user_config_dir, 'settings.yml')

    def __init__(self):
        super(settingstracker, self).__init__(name='Settings')

        try:
            with open(self.settingspath,'r') as stream:
                self.restoreState(yaml.load(stream))
            for param in self.template()['children']:
                if param['name'] not in self:
                    raise yaml.YAMLError
        except (yaml.YAMLError,IOError) as exc:
            msg.logMessage(exc, msg.WARNING)
            self.restoreState(self.template())
        self.sigTreeStateChanged.connect(self.write)


    def write(self):

        if not os.path.exists(pathtools.user_config_dir):
            os.makedirs(pathtools.user_config_dir)
        with open(self.settingspath,'w') as stream:
            try:
                stream.write(yaml.dump(self.saveState()))
                stream.close()
            except yaml.YAMLError as exc:
                print exc

    def __getitem__(self, item):
        if item in self:
            try:
                return super(settingstracker, self).__getitem__(item)
            except KeyError:
                return None

    def __setitem__(self, key, value):
        if key in self:
            super(settingstracker, self).__setitem__(key,value)
        else:
            self.addChild({'name':key,'type':type(value).__name__,'value':value})
        self.write()

    def __contains__(self, item):
        return item in self.names

    def _builddialog(self):
        if not hasattr(self,'dialog'):
            self.pt = ParameterTree()
            self.pt.setParameters(self, showTop=False)
            self.dialog = QtGui.QDialog()
            layout = QtGui.QVBoxLayout()
            layout.addWidget(self.pt)
            layout.setContentsMargins(0, 0, 0, 0)
            self.dialog.setLayout(layout)
            self.dialog.setWindowTitle('Settings')

    def showEditor(self):
        self._builddialog()
        self.dialog.show()

    # TODO check for new template fields on start
    @staticmethod
    def template():
        return {'type':'group','name':'settings','children':[
            {'name':'Default Local Path','value':os.path.expanduser('~'),'type':'str'},
            {'name':'Integration Bins (q)','value':1000,'type':'int','min':1},
            {'name': 'Integration Bins (χ)', 'value': 1000, 'type': 'int','min':1},
            {'name':'Image Load Rotations','value':0,'type':'int'},
            {'name':'Image Load Transpose','value':False,'type':'bool'},
            {'name':'Ignored Modules','value':[],'type':'list'}]}


settings=settingstracker()

class PyFAIGeometry(pyFAI.geometry.Geometry):
    def set_fit2d(self,
                  wavelength,
                  distance,
                  center_x,
                  center_y,
                  tilt,
                  rotation):
        self.set_wavelength(wavelength * 1e-10)
        self.setFit2D(distance, center_x, center_y, tilt, rotation)

    def get_fit2d(self):
        param_dict = self.getFit2D()
        return [self.get_wavelength() * 1e10,
                param_dict['directDist'],
                param_dict['centerX'],
                param_dict['centerY'],
                param_dict['tilt'],
                param_dict['tiltPlanRotation']]



class experiment(Parameter):
    def __init__(self, path=None):

        self.imageshape = (1475, 1709)

        if path is None:  # If not loading an exeriment from file
            # Build an empty experiment tree
            config = [{'name': 'Name', 'type': 'str', 'value': 'New Experiment'},
                      {'name': 'Detector', 'type': 'list', 'values':detectors.ALL_DETECTORS},
                      {'name': 'Pixel Size X', 'type': 'float', 'value': 172.e-6, 'siPrefix': True, 'suffix': 'm',
                       'step': 1e-6},
                      {'name': 'Pixel Size Y', 'type': 'float', 'value': 172.e-6, 'siPrefix': True, 'suffix': 'm',
                       'step': 1e-6},
                      {'name': 'Center X', 'type': 'float', 'value': 0, 'suffix': ' px'},
                      {'name': 'Center Y', 'type': 'float', 'value': 0, 'suffix': ' px'},
                      {'name': 'Detector Distance', 'type': 'float', 'value': 1, 'siPrefix': True, 'suffix': 'm',
                       'step': 1e-3},
                      {'name': 'Detector Tilt', 'type': 'float', 'value': 0, 'siPrefix': False, 'suffix': u'°',
                       'step': 1e-1},
                      {'name': 'Detector Rotation', 'type': 'float', 'value': 0, 'siPrefix': False, 'suffix': u'°',
                       'step': 1e-1},
                      {'name': 'Energy', 'type': 'float', 'value': 10000, 'siPrefix': True, 'suffix': 'eV'},
                      {'name': 'Wavelength', 'type': 'float', 'value': 1, 'siPrefix': True, 'suffix': 'm'},
                      # {'name': 'View Mask', 'type': 'action'},
                      {'name': 'Incidence Angle (GIXS)', 'type': 'float', 'value': 0.1, 'suffix': u'°'},
                      {'name': 'Notes', 'type': 'text', 'value': ''}]
            super(experiment, self).__init__(name='Experiment Properties', type='group', children=config)

            # Wire up the energy and wavelength parameters to fire events on change (so they always match)
            self.param('Energy').sigValueChanged.connect(self.EnergyChanged)
            self.param('Wavelength').sigValueChanged.connect(self.WavelengthChanged)
            self.param('Detector').sigValueChanged.connect(self.DetectorChanged)

            # Add tilt style dialog
            self.tiltStyleMenu = QtGui.QMenu()
            grp = QtGui.QActionGroup(self)
            self.fit2dstyle = QtGui.QAction('Use Fit2D style rot/tilt', self.tiltStyleMenu)
            self.wxdiffstyle = QtGui.QAction('Use WxDiff style rot/tilt', self.tiltStyleMenu)
            self.fit2dstyle.setCheckable(True)
            self.wxdiffstyle.setCheckable(True)
            self.fit2dstyle.setActionGroup(grp)
            self.wxdiffstyle.setActionGroup(grp)
            self.fit2dstyle.setChecked(True)
            self.fit2dstyle.triggered.connect(self.resetUnits)
            self.wxdiffstyle.triggered.connect(self.resetUnits)

            self.tiltStyleMenu.addActions([self.fit2dstyle, self.wxdiffstyle])

            # Start with a null mask
            self._mask = None

            self.EnergyChanged()
        else:
            # Load the experiment from file
            with open(path, 'r') as f:
                self.config = pickle.load(f)

        self.headermap = {'Beam Energy': 'Beam Energy',
                          'Sample Alpha Stage': 'Sample Alpha Stage',
                          'Detector Vertical': 'Detector Vertical',
                          'Detector Horizontal': 'Detector Horizontal',
                          'I1 AI': 'I1 AI',
                          'Timeline Axis': None}

    # Make the mask accessible as a property
    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, value):
        self._mask = value

    @mask.deleter
    def mask(self):
        del self._mask

    def addtomask(self, maskedarea):
        # If the mask is empty, set the mask to the new masked area
        if self._mask is None or self._mask.shape != maskedarea.shape:
            self._mask = maskedarea.astype(np.int)
        else:  # Otherwise, bitwise or it with the current mask
            # print(self.experiment.mask,maskedarea)
            self._mask = np.bitwise_and(self._mask, maskedarea.astype(np.int))

        # try:  # hack for mask export
        #     from fabio import edfimage
        #
        #     edf = edfimage.edfimage(np.rot90(self.mask))
        #     edf.write('mask.edf')
        # except Exception:
        #     pass

    def DetectorChanged(self):
        self.param('Pixel Size X').setValue(self.getDetector().get_pixel1())
        self.param('Pixel Size Y').setValue(self.getDetector().get_pixel2())

    def EnergyChanged(self):
        # Make Energy and Wavelength match
        self.param('Wavelength').setValue(1.239842e-6 / self.param('Energy').value(),
                                          blockSignal=self.WavelengthChanged)

    def WavelengthChanged(self):
        # Make Energy and Wavelength match
        self.param('Energy').setValue(1.239842e-6 / self.param('Wavelength').value(), blockSignal=self.EnergyChanged)

    def save(self):
        # Save the experiment .....
        with open(self.getvalue('Name') + '.exp', 'w') as f:
            pickle.dump(self.saveState(), f)
        with open(self.getvalue('Name') + '.expmask', 'w') as f:
            np.save(f, self.mask)

    def getvalue(self, name):
        # Return the value of the named child
        return self.child(name).value()

    def setvalue(self, name, value):
        # Set the value of the named child
        self.child(name).setValue(value)

    @property
    def center(self):
        return self.getvalue('Center X'), self.getvalue('Center Y')

    @center.setter
    def center(self, cen):
        self.setvalue('Center X', cen[0])
        self.setvalue('Center Y', cen[1])

    def getAI(self):
        """
        :rtype : pyFAI.AzimuthalIntegrator
        """
        # print(self.getDetector().MAX_SHAPE)
        AI = pyFAI.AzimuthalIntegrator(
            wavelength=self.getvalue('Wavelength'))
        #                                dist=self.getvalue('Detector Distance'),
        #                                poni1=self.getvalue('Pixel Size X') * (self.getvalue('Center Y')),
        #                                poni2=self.getvalue('Pixel Size Y') * (self.getvalue('Center X')),
        #                                rot1=0,
        #                                rot2=0,
        #                                rot3=0,
        #                                pixel1=self.getvalue('Pixel Size Y'),
        #                                pixel2=self.getvalue('Pixel Size X'),
        #                                detector=self.getDetector(),
        if self.fit2dstyle.isChecked():
            AI.setFit2D(self.getvalue('Detector Distance') * 1000.,
                        self.getvalue('Center X'),
                        self.getvalue('Center Y'),
                        self.getvalue('Detector Tilt'),
                        360. - self.getvalue('Detector Rotation'),
                        self.getvalue('Pixel Size Y') * 1.e6,
                        self.getvalue('Pixel Size X') * 1.e6)
        elif self.wxdiffstyle.isChecked():
            AI.setFit2D(self.getvalue('Detector Distance') * 1000.,
                        self.getvalue('Center X'),
                        self.getvalue('Center Y'),
                        self.getvalue('Detector Tilt') / 2. / np.pi * 360.,
                        360. - (2 * np.pi - self.getvalue('Detector Rotation')) / 2. / np.pi * 360.,
                        self.getvalue('Pixel Size Y') * 1.e6,
                        self.getvalue('Pixel Size X') * 1.e6)
        AI.set_wavelength(self.getvalue('Wavelength'))
        # print AI
        return AI

        # def getGeometry(self):
        #     """
        #     :rtype : pyFAI.Geometry
        #     """
        #     # print(self.getDetector().MAX_SHAPE)
        #     geo = PyFAIGeometry(dist=self.getvalue('Detector Distance'),
        #                         poni1=self.getvalue('Pixel Size X') * (self.getvalue('Center Y')),
        #                         poni2=self.getvalue('Pixel Size Y') * (self.getvalue('Center X')),
        #                         rot1=0,
        #                         rot2=0,
        #                         rot3=0,
        #                         pixel1=self.getvalue('Pixel Size Y'),
        #                         pixel2=self.getvalue('Pixel Size X'),
        #                         detector=self.getDetector(),
        #                         wavelength=self.getvalue('Wavelength'))
        # geo = PyFAIGeometry(wavelength=self.getvalue('Wavelength'))
        # geo.setFit2D(self.getvalue('Detector Distance'),
        #             self.getvalue('Center Y'),
        #             self.getvalue('Center X'),
        #             self.getvalue('Detector Tilt'),
        #             360.-self.getvalue('Detector Rotation'),
        #             self.getvalue('Pixel Size Y')*1.e6,
        #             self.getvalue('Pixel Size X')*1.e6)
        # print AI

        # return geo

    def getDetector(self):
        return self.getvalue('Detector')()

    def edit(self):
        pass



        # edit the data
        # config['key3'] = 'value3'

        # write it back to the file
        # with open('config.json', 'w') as f:
        #    json.dump(config, f)

    def iscalibrated(self):
        return (self.getvalue('Pixel Size X') > 0) and (self.getvalue('Pixel Size Y') > 0) and (
            self.getvalue('Detector Distance') > 0)

    def setHeaderMap(self, xikey, headerkey):
        self.headermap[xikey] = headerkey

    def mapHeader(self, xikey):
        return self.headermap[xikey]

    @property
    def qcorners(self):
        pass
        # bl=(0,0)
        # br=(self.imageshape[0],0)
        # tl=(0,self.imageshape[1])
        # tr=self.imageshape
        # self.getGeometry()
        # from pyFAI import geometry
        # l=geometry.Geometry.qFunction()
        #     qFunction(d1, d2, param=None, path='cython')

    def resetUnits(self):
        tilt = self.param('Detector Tilt')
        rot = self.param('Detector Rotation')
        if self.fit2dstyle.isChecked():
            tilt.setOpts(**{'suffix': u'°'})
            rot.setOpts(**{'suffix': u'°'})
        elif self.wxdiffstyle.isChecked():
            rot.setOpts(**{'suffix': ' rad'})
            tilt.setOpts(**{'suffix': ' rad'})


activeExperiment = None


def activate():
    global activeExperiment
    activeExperiment = experiment()



import pyqtgraph as pg
from PySide import QtGui, QtCore
import numpy as np


class QCircRectF(QtCore.QRectF):
    def __init__(self, center=(0, 0), radius=1, rect=None):
        self._scale = 1.
        if rect is not None:
            self.center = rect.center()
            super(QCircRectF, self).__init__(rect)
        else:
            self.center = QtCore.QPointF(*center)

            left = self.center.x() - radius
            top = self.center.y() - radius
            bottom = self.center.y() + radius
            right = self.center.x() + radius

            super(QCircRectF, self).__init__(QtCore.QPointF(left, top), QtCore.QPointF(right, bottom))

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, value):
        self._scale = value
        self.radius *= value
        self.setLeft(self.center.x() - self._radius)
        self.setTop(self.center.y() - self._radius)
        self.setBottom(self.center.y() + self._radius)
        self.setRight(self.center.x() + self._radius)

    @property
    def radius(self):
        return (self.right() - self.left()) * .5

    @radius.setter
    def radius(self, radius):

        self.setLeft(self.center.x() - radius)
        self.setTop(self.center.y() - radius)
        self.setBottom(self.center.y() + radius)
        self.setRight(self.center.x() + radius)



class QRectF(QtCore.QRectF):
    def scale(self, ratio):
        coords = [coord * ratio for coord in self.getCoords()]

        self.setCoords(*coords)


class LinearRegionItem(pg.LinearRegionItem):
    sigRemoveRequested = QtCore.Signal(object)

    def __init__(self, *args, **kwargs):
        super(LinearRegionItem, self).__init__(*args, **kwargs)
        self.menu = None
        self.isdeleting = False

    def getArrayRegion(self,data,item): #rotate once?
        cut=np.zeros_like(data)
        if self.orientation is pg.LinearRegionItem.Horizontal:
            regionbounds = self.getRegion()

            cut[:, int(regionbounds[0]):int(regionbounds[1])] = 1
        elif self.orientation is pg.LinearRegionItem.Vertical:
            regionbounds = self.getRegion()

            cut[int(regionbounds[0]):int(regionbounds[1]), :] = 1


        return (cut * data).T

    def hoverEvent(self, ev):
        hover = False
        if not ev.isExit():
            if ev.acceptDrags(QtCore.Qt.LeftButton):
                hover = True

            for btn in [QtCore.Qt.LeftButton, QtCore.Qt.RightButton, QtCore.Qt.MidButton]:
                if int(self.acceptedMouseButtons() & btn) > 0 and ev.acceptClicks(btn):
                    hover = True
            ev.acceptClicks(QtCore.Qt.RightButton)

        if hover:
            self.setMouseHover(True)

            ev.acceptClicks(
                QtCore.Qt.LeftButton)  ## If the ROI is hilighted, we should accept all clicks to avoid confusion.
            ev.acceptClicks(QtCore.Qt.RightButton)
            ev.acceptClicks(QtCore.Qt.MidButton)
        else:
            self.setMouseHover(False)

    def raiseContextMenu(self, ev):
        menu = self.getMenu()
        menu = self.scene().addParentContextMenus(self, menu, ev)
        pos = ev.screenPos()
        menu.popup(QtCore.QPoint(pos.x(), pos.y()))


    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.RightButton:
            self.raiseContextMenu(ev)
            ev.accept()
        elif int(ev.button() & self.acceptedMouseButtons()) > 0:
            ev.accept()
            self.sigClicked.emit(self, ev)
        else:
            ev.ignore()

    def getMenu(self):
        if self.menu is None:
            self.menu = QtGui.QMenu()
            self.menu.setTitle("ROI")
            remAct = QtGui.QAction("Remove ROI", self.menu)
            remAct.triggered.connect(self.removeClicked)
            self.menu.addAction(remAct)
            self.menu.remAct = remAct
        return self.menu

    def removeClicked(self):
        self.isdeleting = True
        QtCore.QTimer.singleShot(0, lambda: self.sigRemoveRequested.emit(self))
        # self.deleteLater()
        ## Send remove event only after we have exited the menu event handler
        #QtCore.QTimer.singleShot(0, lambda: self.sigRemoveRequested.emit(self))




class LineROI(pg.LineROI):
    def __init__(self, *args, **kwargs):
        super(LineROI, self).__init__(*args, **kwargs)
        #self.sigRemoveRequested.connect(self.deleteLater)
        self.isdeleting = False
        self.sigRemoveRequested.connect(self.delete)

    def getArrayRegion(self, data, img, axes=(0, 1), returnMappedCoords=False, **kwds):
        from skimage.draw import polygon

        shape, v, origin = self.getAffineSliceParams(data, img, axes)

        br = np.add(origin, np.multiply(v[0], shape[0]))
        tl = np.add(origin, np.multiply(v[1], shape[1]))
        bl = origin
        tr = np.sum([origin, np.multiply(v[1], shape[1]), np.multiply(v[0], shape[0])], axis=0)
        vecs = np.vstack([bl, br, tr, tl]).T
        rr, cc = polygon(vecs[0], vecs[1])
        mask = np.zeros_like(data)
        rrcc = [[r, c] for r, c in zip(rr, cc) if r < mask.shape[0] and c < mask.shape[1]]
        rr, cc = zip(*rrcc)
        mask[rr, cc] = 1

        return (data * mask).T.copy()

    def delete(self):
        self.isdeleting = True

class ArcROI(pg.ROI):
    """
    Elliptical ROI subclass with one scale handle and one rotation handle.


    ============== =============================================================
    **Arguments**
    pos            (length-2 sequence) The position of the ROI's origin.
    size           (length-2 sequence) The size of the ROI's bounding rectangle.
    \**args        All extra keyword arguments are passed to ROI()
    ============== =============================================================

    """

    #sigRemoveRequested = QtCore.Signal()

    def __init__(self, center, radius, **args):
        # QtGui.QGraphicsRectItem.__init__(self, 0, 0, size[0], size[1])
        r = QCircRectF(center, radius)
        super(ArcROI, self).__init__(r.center, r.size(), removable=True, **args)
        #self.addRotateHandle([1.0, 0.5], [0.5, 0.5])
        #self.addScaleHandle([0.5*2.**-0.5 + 0.5, 0.5*2.**-0.5 + 0.5], [0.5, 0.5])

        self.innerhandle = self.addFreeHandle([0., .25])
        self.outerhandle = self.addFreeHandle([0., .5])
        self.lefthandle = self.addFreeHandle([.433, .25])
        self.righthandle = self.addFreeHandle([-.433, .25])

        self.startangle = 30
        self.arclength = 120

        self.aspectLocked = True
        self.translatable = False
        self.translateSnap = False
        self.removable = True

        self.cacheinner = self.innerhandle.pos()
        self.cacheouter = self.outerhandle.pos()
        self.startradius = radius
        self.startcenter = center

        self.isdeleting = False
        self.sigRemoveRequested.connect(self.delete)


    def getRadius(self):
        radius = pg.Point(self.outerhandle.pos()).length()
        # r2 = radius[1]
        #r3 = r2[0]
        return radius

    def getInnerRadius(self):
        radius = pg.Point(self.innerhandle.pos()).length()
        # r2 = radius[1]
        #r3 = r2[0]
        return radius / self.getRadius() * .5

    def boundingRect(self):
        r = self.getRadius()
        return QtCore.QRectF(-r * 1., -r * 1., 2.0 * r, 2.0 * r)

    def getCenter(self):
        r = self.boundingRect()
        r = QCircRectF(rect=r)
        return r.center
    

    def paint(self, p, opt, widget):
        pen = QtGui.QPen()
        pen.setColor(QtGui.QColor(0, 255, 255))

        p.setPen(pen)

        # if self.cacheouter!=self.outerhandle.pos():
        #    self.innerhandle.setPos(self.cacheinner)


        r = self.boundingRect()
        #p.drawRect(r)
        p.setRenderHint(QtGui.QPainter.Antialiasing)

        p.scale(r.width(), r.height())  ## workaround for GL bug

        #r = QRectF(r.x() / r.width(), r.y() / r.height(), 1., 1.)


        # p.drawEllipse(r)
        r = QCircRectF(radius=0.5)
        #r.radius=self.getRadius()


        self.arclength = np.degrees(np.arctan2(self.lefthandle.pos().x(), self.lefthandle.pos().y()) -
                                    np.arctan2(self.righthandle.pos().x(), self.righthandle.pos().y()))
        self.startangle = np.degrees(np.arctan2(self.lefthandle.pos().y(), self.lefthandle.pos().x()))
        # print 'start: ',startangle
        #print 'length: ',arclength


        p.drawArc(r, -(self.startangle) * 16, -(self.arclength) * 16)


        # pos = self.mapFromView(self.innerhandle.pos())
        radius = self.getInnerRadius()

        r = QCircRectF()

        r.radius = radius
        #p.drawRect(r)
        #print r.radius
        p.drawArc(r, -(self.startangle) * 16, -self.arclength * 16)

        #p.drawLine(self.lefthandle.pos().norm()*self.innerhandle.pos().length()/2,self.lefthandle.pos().norm()/2)
        #p.drawLine(self.righthandle.pos().norm()*self.innerhandle.pos().length()/2.,self.righthandle.pos().norm()/2)



        pen.setStyle(QtCore.Qt.DashLine)
        p.setPen(pen)

        p.drawLine(QtCore.QPointF(0., 0.), self.lefthandle.pos() * 100)
        p.drawLine(QtCore.QPointF(0., 0.), self.righthandle.pos() *100)




    def getArrayRegion(self, arr, img=None):
        """
        Return the result of ROI.getArrayRegion() masked by the elliptical shape
        of the ROI. Regions outside the ellipse are set to 0.
        """
        # arr = pg.ROI.getArrayRegion(self, arr, img)
        #if arr is None or arr.shape[0] == 0 or arr.shape[1] == 0:
        #    return None
        w = arr.shape[0]
        h = arr.shape[1]
        ## generate an ellipsoidal mask
        mask = np.fromfunction(
            lambda x, y: (self.innerhandle.pos().length() < (
            (x - self.startcenter[0]) ** 2. + (y - self.startcenter[1]) ** 2.) ** .5) &
                         (((x - self.startcenter[0]) ** 2. + (
                         y - self.startcenter[1]) ** 2.) ** .5 < self.outerhandle.pos().length()) &
                         ((np.degrees(np.arctan2(y - self.startcenter[1], x - self.startcenter[0]))-self.startangle) %360 > 0) &
                         ((np.degrees(np.arctan2(y - self.startcenter[1],
                                                x - self.startcenter[0]))-self.startangle) %360 < self.arclength)
            , (w, h))

        return (arr * mask).T

    def shape(self):
        self.path = QtGui.QPainterPath()
        self.path.addEllipse(self.boundingRect())
        return self.path

    def delete(self):
        self.isdeleting = True
        #self.deleteLater()


from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals
import os
from functools import partial
from collections import OrderedDict

from PySide import QtGui,QtCore,QtUiTools

from paws.qt.qtapi import QPawsAPI
from paws.qt import widgets
from paws.core.models.ListModel import ListModel
from .. import base
from pipeline import msg
from xicam import config
from xicam import threads
from pyqtgraph import parametertree as pt

class BatchPlugin(base.plugin):
    name = 'Batch'

    def __init__(self, *args, **kwargs):
        
        # the dict below specifies available batch workflows for XICAM.
        # note that each batch workflow is composed of two workflows. 
        # one is a processing workflow (written for a single input)
        # and the other is a batch executor which iterates an input 
        # for the processing workflow and optionally harvests its outputs.
        # the names of these two workflows (taken from the paws documentation)
        # are also saved here for later use in paws api calls.
        # depending on what sort of data must be taken from Xi-cam internal variables,
        # each workflow may also perform some additional loading in self.run_wf().
        # - uris to locate workflow modules from paws:
        self._wf_uris = OrderedDict()
        self._wf_uris['saxs integrator'] = 'XICAM.batch_saxs_integrator'
        self._wf_uris['saxs guinier-porod fitter'] = 'XICAM.batch_saxs_gp_fit'
        # - names of the processing workflows:
        self._wf_names = OrderedDict()
        self._wf_names['saxs integrator'] = 'saxs_integrator'
        self._wf_names['saxs guinier-porod fitter'] = 'saxs_gp_fit'
        # - names of the batch-execution workflows:
        self._batch_wf_names = OrderedDict()
        self._batch_wf_names['saxs integrator'] = 'batch_saxs_integrator'
        self._batch_wf_names['saxs guinier-porod fitter'] = 'batch_saxs_gp_fit'
        self._run_button_text = '&Run'
        self._stop_button_text = 'S&top'

        # create basic ui components
        self.build_ui()
        self.centerwidget = self.viewer_frame
        self.rightwidget = self.wf_frame
        self.bottomwidget = self.batch_control

        # PawsAPI instances for each workflow,
        # and ParameterTree roots as well
        self._paws = OrderedDict()
        self._root_params = OrderedDict()
        #self._run_wf_buttons = OrderedDict()
        # allowed datatypes for ParameterTree entries:
        self.allowed_datatypes = ['int','float','str','bool']
        for wf_title,wf_uri in self._wf_uris.items():
            self._paws[wf_title] = QPawsAPI(QtGui.QApplication.instance())
            self._paws[wf_title].load_packaged_workflow(wf_uri)
            self._add_workflow_params(wf_title)
            # TODO : How should message emissions be handled?
            self._paws[wf_title].set_logmethod(print)
            self._paws[wf_title]._wf_manager.emitMessage.connect(print)
            # connect the signal/slot to update the "Run" button
            self._paws[wf_title].get_wf(self._batch_wf_names[wf_title]).wfFinished.connect( partial(self._reset_run_wf_button,wf_title) )
            # Connect the viewer
            #self._paws[wf_title].get_wf(self._wf_names[wf_title]).opFinished.connect( self.update_visuals )

        # select a default workflow (populates ui content)
        self.select_workflow(self._wf_uris.keys()[0])

        #self._current_visual = None
        self._vis_widget = None
        self._placeholder_widget = QtGui.QPlainTextEdit('intentionally blank')

        super(BatchPlugin, self).__init__(*args, **kwargs)

    def build_ui(self):
        self.remove_files_button = QtGui.QPushButton('Remove selected files')
        self.remove_files_button.clicked.connect(self.rm_files)

        self.viewer_frame = QtGui.QFrame()
        self.viewer_layout = QtGui.QGridLayout()
        self.viewer_frame.setLayout(self.viewer_layout)

        self.wf_frame = QtGui.QFrame()
        self.wf_layout = QtGui.QGridLayout()
        self.wf_frame.setLayout(self.wf_layout)
        self.wf_selector = QtGui.QComboBox()
        lm = ListModel(self._wf_uris.keys())
        self.wf_selector.setModel(lm)
        self.wf_selector.currentIndexChanged.connect( partial(self._set_wf_by_idx) )
        self.run_wf_button = QtGui.QPushButton(self._run_button_text)
        self.run_wf_button.clicked.connect(self.toggle_run_wf)
        self.vis_toggle = QtGui.QCheckBox('Realtime visualization')
        self.wf_control = pt.ParameterTree()
        self.wf_control.setHeaderLabels(['Operation','Settings'])
        self.wf_layout.addWidget(self.wf_selector,0,0,1,2)
        self.wf_layout.addWidget(self.run_wf_button,1,0,1,1)
        self.wf_layout.addWidget(self.vis_toggle,1,1,1,1)
        self.wf_layout.addWidget(self.wf_control,2,0,1,2)
        #self.wf_control.itemClicked.connect(self._display_wf_item)

        self.batch_control = QtGui.QWidget()
        self.batch_list = QtGui.QListWidget()
        self.batch_layout = QtGui.QGridLayout()
        self.batch_control.setLayout(self.batch_layout)
        self.batch_layout.addWidget(self.batch_list,0,0,1,1)
        self.batch_layout.addWidget(self.remove_files_button,1,0,1,1)
        self.batch_list.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)

    def select_workflow(self,wf_title): 
        self.wf_control.setParameters(
            self._root_params[wf_title],showTop=False)
        self._reset_run_wf_button(wf_title)

    # this is an overridden method from base.plugin 
    def openfiles(self, files, operation=None, operationname=None):
        self.batch_list.addItems(files)

    def rm_files(self):
        itms = self.batch_list.selectedItems()
        for itm in itms:
            self.batch_list.takeItem(self.batch_list.row(itm))
 
    def _rt_vis_active(self):
        return self.vis_toggle.isChecked()

    def _current_wf_title(self):
        wf_idx = self.wf_selector.currentIndex()
        return self._wf_uris.keys()[wf_idx]

    def _set_wf_by_idx(self,wf_selector_idx):
        wf_title = self.wf_selector.model().list_data()[wf_selector_idx]
        self.select_workflow(wf_title)

    def _add_workflow_params(self,wf_title):
        paw = self._paws[wf_title]
        # Create a root parameter for the ParameterTree
        root_param = pt.types.GroupParameter(name='root')
        # Create child parameters for all operations in the processing workflow
        for op_tag in paw.get_wf(self._wf_names[wf_title]).list_op_tags():
            # Connect the ParameterTree
            root_param.addChild(self._op_param(op_tag,wf_title))
        self._root_params[wf_title] = root_param
        # Create a "run" button as an "ActionParameter"
        #run_wf_button=pt.types.ActionParameter(name=self._run_button_text)
        #run_wf_button.sigActivated.connect( partial(self.toggle_run_wf,wf_title) )
        #root_param.addChild(run_wf_button)
        #self._run_wf_buttons[wf_title] = run_wf_button

    def _op_param(self,op_tag,wf_title):
        paw = self._paws[wf_title]
        # op Parameter 
        p = pt.types.SimpleParameter(   
            name=op_tag, 
            type='bool',
            showTop=False, 
            value=paw.is_op_enabled(op_tag,self._wf_names[wf_title]), 
            expanded=False)
        p.sigValueChanged.connect( partial(self._set_op_enabled,wf_title,op_tag) )
        # child Parameters: inputs and outputs
        op = paw.get_op(op_tag,self._wf_names[wf_title])
        for inp_name in op.inputs.keys():
            inp_val = paw.get_input_setting(op_tag,inp_name,self._wf_names[wf_title]) 
            if op.input_datatype[inp_name] in self.allowed_datatypes:
                p_inp = pt.types.SimpleParameter(
                    name=inp_name,
                    type=op.input_datatype[inp_name],
                    value=inp_val)
            else:
                p_inp = pt.types.SimpleParameter(
                    name=inp_name,
                    type='str',
                    value=str(inp_val))
            p_inp.sigValueChanged.connect( partial(self._set_parameter,wf_title,op_tag,inp_name) )
            p.addChild(p_inp)
            # TODO: Connect paws opChanged signal to some slot 
            # that updates parameters in self.wf_control 
            # for parameters that refer to workflow item inputs 
        for out_name in op.outputs.keys():
            p_out = pt.types.ActionParameter(name='display '+out_name)
            p_out.sigActivated.connect( partial(self._display_op_output,wf_title,op_tag,out_name) )
            p.addChild(p_out)
        return p

    def _display_op_output(self,wf_title,op_name,out_name,param):
        paw = self._paws[wf_title]
        wf = paw.get_wf(self._wf_names[wf_title])
        data = wf.get_op_output(op_name,out_name)
        if data is not None:
            widg = widgets.make_widget(data)
        else:
            widg = self._placeholder_widget
        # clear the output layout
        old_widg_itm = self.viewer_layout.takeAt(0)
        if old_widg_itm is not None:
            if old_widg_itm.widget() is not self._placeholder_widget:
                old_widg_itm.widget().close()
            del old_widg_itm
            #old_widg_itm.widget().deleteLater()
        # set the new widget
        self.viewer_layout.addWidget(widg,0,0)
        #self._current_visual = op_name
        self._vis_widget = widg

    def _set_parameter(self,wf_title,op_name,input_name,param,val):
        paw = self._paws[wf_title]
        paw.set_input(op_name,input_name,val,None,self._wf_names[wf_title])

    def _set_op_enabled(self,wf_title,op_name,param,val):
        paw = self._paws[wf_title]
        if val:
            paw.enable_op(op_name,self._wf_names[wf_title])
        else:
            paw.disable_op(op_name,self._wf_names[wf_title])

    def _reset_run_wf_button(self,wf_title):
        if wf_title == self._current_wf_title():
            wf_run_status = self._paws[wf_title]._wf_manager.wf_running[self._batch_wf_names[wf_title]]
            if wf_run_status:
                self.run_wf_button.setText(self._stop_button_text)
            else:
                self.run_wf_button.setText(self._run_button_text)

    def toggle_run_wf(self):
        wf_title = self._current_wf_title()
        paw = self._paws[wf_title]
        wf_run_status = paw._wf_manager.wf_running[self._batch_wf_names[wf_title]]
        if wf_run_status:
            self.stop_wf(wf_title)
        else:
            self.run_wf(wf_title)

    def stop_wf(self,wf_title):
        self._paws[wf_title].stop_wf(self._batch_wf_names[wf_title])
        self._reset_run_wf_button(wf_title)

    def run_wf(self,wf_title):
        paw = self._paws[wf_title]
        if wf_title == self._current_wf_title():
            self.run_wf_button.setText(self._stop_button_text)

        # harvest file list and (maybe) PyFAI.AzimuthalIntegrator settings
        # from GUI and Xi-cam internal variables, respectively
        file_list = []
        nfiles = self.batch_list.count()
        for r in range(nfiles):
            p = self.batch_list.item(r).text()
            file_list.append(p)
        paw.set_input('Batch Execution','input_arrays',[file_list],
            None,self._batch_wf_names[wf_title])
        if wf_title in ['saxs integrator']:
            paw.set_input('Integrator Setup','poni_dict',
                config.activeExperiment.getAI().getPyFAI(),
                None,self._batch_wf_names[wf_title])

        #if wfmanager.client is not None:
        #    wfmanager.run_paws(paw)
        #else:
        run_off_thread = threads.method()( paw.execute )
        run_off_thread( self._batch_wf_names[wf_title] )

    #def update_visuals(self,op_tag):
    #    if op_tag == self._current_visual:
    #        if self._vis_widget is self._placeholder_widget:
    #            # visual was previously not ready: try again
    #            self.set_visualizer(op_tag)
    #        else:
    #            visdata = self._get_vis_output(op_tag)
    #            if op_tag in ['Read Image','Integrate to 2d','log(I) 2d']: 
    #                # expect a pyqtgraph.ImageView
    #                self._vis_widget.setImage(visdata) 
    #            elif op_tag in ['Integrate to 1d','log(I) 1d']:
    #                # expect a pyqtgraph.PlotWidget
    #                self._vis_widget.getPlotItem().clear()
    #                self._vis_widget.getPlotItem().plot(visdata)
    #            elif op_tag in ['Output CSV','Output Image']:
    #                # expect a QtGui.QTextEdit
    #                t = widgets.display_text_fast(visdata)
    #                self._vis_widget.setText(t)
    #    self.paw.app.processEvents()


